# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

__all__ = ["CombinedScanTask", "DataFile", "DataOperations", "FileFormat", "FileScanTask",
           "Files", "Filterable", "FilteredSnapshot", "ManifestFile", "PartitionFieldSummary",
           "Metrics", "PartitionSpec", "PartitionSpecBuilder",
           "Schema", "Snapshot", "SnapshotIterable", "StructLike",
           "Table", "Tables", "TableScan", "Transaction", "UpdateSchema"]

from .combined_scan_task import CombinedScanTask
from .data_file import DataFile
from .data_operations import DataOperations
from .file_format import FileFormat
from .file_scan_task import FileScanTask
from .files import Files
from .filterable import Filterable
from .filtered_snapshot import FilteredSnapshot
from .manifest_file import ManifestFile, PartitionFieldSummary
from .metrics import Metrics
from .partition_spec import (PartitionSpec,
                             PartitionSpecBuilder)
from .schema import Schema
from .snapshot import Snapshot
from .snapshot_iterable import SnapshotIterable
from .struct_like import StructLike
from .table import Table
from .table_scan import TableScan
from .tables import Tables
from .transaction import Transaction
from .update_schema import UpdateSchema



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import itertools
import logging
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool

from iceberg.api.expressions import (InclusiveManifestEvaluator,
                                     ResidualEvaluator)

from .base_file_scan_task import BaseFileScanTask
from .base_table_scan import BaseTableScan
from .manifest_reader import ManifestReader
from .partition_spec_parser import PartitionSpecParser
from .schema_parser import SchemaParser
from .table_properties import TableProperties
from .util import SCAN_THREAD_POOL_ENABLED, WORKER_THREAD_POOL_SIZE_PROP


_logger = logging.getLogger(__name__)


class DataTableScan(BaseTableScan):

    SNAPSHOT_COLUMNS = ("snapshot_id", "file_path", "file_ordinal", "file_format", "block_size_in_bytes",
                        "file_size_in_bytes", "record_count", "partition", "value_counts", "null_value_counts",
                        "lower_bounds", "upper_bounds")

    def __init__(self, ops, table, schema=None, snapshot_id=None, row_filter=None,
                 case_sensitive=True, selected_columns=None, options=None, minused_cols=None):
        super(DataTableScan, self).__init__(ops, table, schema if schema is not None else table.schema(),
                                            snapshot_id=snapshot_id, row_filter=row_filter,
                                            case_sensitive=case_sensitive, selected_columns=selected_columns,
                                            options=options, minused_cols=minused_cols)
        self._cached_evaluators = dict()

    def new_refined_scan(self, ops, table, schema, snapshot_id=None, row_filter=None, case_sensitive=None,
                         selected_columns=None, options=None, minused_cols=None):
        return DataTableScan(ops, table, schema,
                             snapshot_id=snapshot_id, row_filter=row_filter, case_sensitive=case_sensitive,
                             selected_columns=selected_columns, options=options, minused_cols=minused_cols)

    def plan_files(self, ops=None, snapshot=None, row_filter=None):
        if all(i is None for i in [ops, snapshot, row_filter]):
            return super(DataTableScan, self).plan_files()

        matching_manifests = [manifest for manifest in snapshot.manifests
                              if self.cache_loader(manifest.spec_id).eval(manifest)]

        if self.ops.conf.get(SCAN_THREAD_POOL_ENABLED):
            with Pool(self.ops.conf.get(WORKER_THREAD_POOL_SIZE_PROP,
                                        cpu_count())) as reader_scan_pool:
                return itertools.chain.from_iterable([scan for scan
                                                      in reader_scan_pool.map(self.get_scans_for_manifest,
                                                                              matching_manifests)])
        else:
            return itertools.chain.from_iterable([self.get_scans_for_manifest(manifest)
                                                  for manifest in matching_manifests])

    def cache_loader(self, spec_id):
        spec = self.ops.current().spec_id(spec_id)
        return InclusiveManifestEvaluator(spec, self.row_filter)

    def get_scans_for_manifest(self, manifest):
        from .filesystem import FileSystemInputFile
        input_file = FileSystemInputFile.from_location(manifest.manifest_path, self.ops.conf)
        reader = ManifestReader.read(input_file)
        schema_str = SchemaParser.to_json(reader.spec.schema)
        spec_str = PartitionSpecParser.to_json(reader.spec)
        residuals = ResidualEvaluator(reader.spec, self.row_filter)
        return [BaseFileScanTask(file, schema_str, spec_str, residuals)
                for file in reader.filter_rows(self.row_filter).select(BaseTableScan.SNAPSHOT_COLUMNS).iterator()]

    def target_split_size(self, ops):
        scan_split_size_str = self.options.get(TableProperties.SPLIT_SIZE)

        if scan_split_size_str is not None:
            try:
                return int(scan_split_size_str)
            except ValueError:
                _logger.warning("Invalid %s option: %s" % (TableProperties.SPLIT_SIZE, scan_split_size_str))

        return int(self.ops.current().properties.get(TableProperties.SPLIT_SIZE, TableProperties.SPLIT_SIZE_DEFAULT))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import copy

from iceberg.api import (DataFile,
                         StructLike)
from iceberg.api.types import StructType

from .avro.iceberg_to_avro import IcebergToAvro
from .partition_data import PartitionData


class GenericDataFile(DataFile, StructLike):

    EMPTY_STRUCT_TYPE = StructType.of([])
    EMPTY_PARTITION_DATA = PartitionData(EMPTY_STRUCT_TYPE)

    def __init__(self, file_path, format, file_size_in_bytes, block_size_in_bytes,
                 row_count=None, partition=None, metrics=None):

        self._file_path = file_path
        self._format = format
        self._row_count = row_count
        self._file_size_in_bytes = file_size_in_bytes
        self._block_size_in_bytes = block_size_in_bytes
        self._file_ordinal = None
        self._sort_columns = None

        if partition is None:
            self._partition_data = GenericDataFile.EMPTY_PARTITION_DATA
            self._partition_type = GenericDataFile.EMPTY_PARTITION_DATA.partition_type
        else:
            self._partition_data = partition
            self._partition_type = partition.get_partition_type()
        if metrics is None:
            self._row_count = row_count
            self._column_sizes = None
            self._value_counts = None
            self._null_value_counts = None
            self._lower_bounds = None
            self._upper_bounds = None
        else:
            self._row_count = metrics.row_count
            self._column_sizes = metrics.column_sizes
            self._value_counts = metrics.value_counts
            self._null_value_counts = metrics.null_value_counts
            self._lower_bounds = metrics.lower_bounds
            self._upper_bounds = metrics.upper_bounds

    def partition(self):
        return self._partition_data

    def path(self):
        return self._file_path

    def format(self):
        return self._format

    def record_count(self):
        return self._row_count

    def file_size_in_bytes(self):
        return self._file_size_in_bytes

    def block_size_in_bytes(self):
        return self._block_size_in_bytes

    def file_ordinal(self):
        return self._file_ordinal

    def sort_columns(self):
        return self._sort_columns

    def column_sizes(self):
        return self._column_sizes

    def value_counts(self):
        return self._value_counts

    def null_value_counts(self):
        return self._null_value_counts

    def lower_bounds(self):
        return self._lower_bounds

    def upper_bounds(self):
        return self._upper_bounds

    def copy(self):
        return copy.deepcopy(self)

    @staticmethod
    def get_avro_schema(partition_type):
        return IcebergToAvro.type_to_schema(DataFile.get_type(partition_type), DataFile.__class__.__name__)

    def __repr__(self):
        fields = ["file_path: {}".format(self._file_path),
                  "file_format: {}".format(self._format),
                  "partition: {}".format(self._partition_data),
                  "record_count: {}".format(self._row_count),
                  "file_size_in_bytes: {}".format(self._file_size_in_bytes),
                  "block_size_in_bytes: {}".format(self._block_size_in_bytes),
                  "column_sizes: {}".format(self._column_sizes),
                  "value_counts: {}".format(self._value_counts),
                  "null_value_counts: {}".format(self._null_value_counts),
                  "lower_bounds: {}".format(self._lower_bounds),
                  "upper_bounds: {}".format(self._upper_bounds),
                  ]
        return "GenericDataFile({})".format("\n,".join(fields))

    def __str__(self):
        return self.__repr__()

    def __deepcopy__(self, memodict):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result

        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memodict))

        return result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .schema import Schema
from .types import (BinaryType,
                    BooleanType,
                    IntegerType,
                    ListType,
                    LongType,
                    NestedField,
                    StringType,
                    StructType)


class ManifestFile(object):
    SCHEMA = Schema(NestedField.required(500, "manifest_path", StringType.get()),
                    NestedField.required(501, "manifest_length", LongType.get()),
                    NestedField.required(502, "partition_spec_id", IntegerType.get()),
                    NestedField.optional(503, "added_snapshot_id", LongType.get()),
                    NestedField.optional(504, "added_data_files_count", IntegerType.get()),
                    NestedField.optional(505, "existing_data_files_count", IntegerType.get()),
                    NestedField.optional(506, "deleted_data_files_count", IntegerType.get()),
                    NestedField
                    .optional(507, "partitions",
                              ListType.of_required(508, StructType.of([NestedField.required(509,
                                                                                            "contains_null",
                                                                                            BooleanType.get()),
                                                                       NestedField.optional(510,
                                                                                            "lower_bound",
                                                                                            BinaryType.get()),
                                                                       NestedField.optional(511,
                                                                                            "upper_bound",
                                                                                            BinaryType.get())]))))

    @staticmethod
    def schema():
        return ManifestFile.SCHEMA

    @property
    def added_files_count(self):
        raise NotImplementedError()

    @property
    def existing_files_count(self):
        raise NotImplementedError()

    @property
    def deleted_files_count(self):
        raise NotImplementedError()

    def copy(self):
        raise NotImplementedError()

    def has_added_files(self):
        return self.added_files_count is None or self.added_files_count > 0

    def has_existing_files(self):
        return self.existing_files_count is None or self.existing_files_count > 0

    def has_deleted_files(self):
        return self.deleted_files_count is None or self.deleted_files_count > 0


class PartitionFieldSummary(object):
    TYPE = ManifestFile.schema().find_type("partitions").as_list_type().element_type.as_struct_type()

    @staticmethod
    def get_type():
        return PartitionFieldSummary.TYPE

    def contains_null(self):
        raise NotImplementedError()

    def lower_bound(self):
        raise NotImplementedError()

    def upper_bound(self):
        raise NotImplementedError()

    def copy(self):
        raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class PartitionField(object):

    def __init__(self, source_id, field_id, name, transform):
        self.source_id = source_id
        self.field_id = field_id
        self.name = name
        self.transform = transform

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, PartitionField):
            return False

        return self.source_id == other.source_id and self.field_id == other.field_id and \
            self.name == other.name and self.transform == other.transform

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return PartitionField.__class__, self.source_id, self.field_id, self.name, self.transform



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import threading

from .expressions import Expressions, ExpressionVisitors
from ..expressions.binder import Binder
from ..types import Conversions


class InclusiveMetricsEvaluator(object):

    def __init__(self, schema, unbound, case_sensitive=True):
        self.schema = schema
        self.struct = schema.as_struct()
        self.case_sensitive = case_sensitive
        self.expr = Binder.bind(self.struct, Expressions.rewrite_not(unbound), case_sensitive)
        self.thread_local_data = threading.local()

    def _visitor(self):
        if not hasattr(self.thread_local_data, "visitors"):
            self.thread_local_data.visitors = MetricsEvalVisitor(self.expr, self.schema, self.struct)

        return self.thread_local_data.visitors

    def eval(self, file):
        return self._visitor().eval(file)


class MetricsEvalVisitor(ExpressionVisitors.BoundExpressionVisitor):
    ROWS_MIGHT_MATCH = True
    ROWS_CANNOT_MATCH = False

    def __init__(self, expr, schema, struct):
        self.expr = expr
        self.value_counts = None
        self.null_counts = None
        self.lower_bounds = None
        self.upper_bounds = None
        self.schema = schema
        self.struct = struct

    def eval(self, file):
        if file.record_count() <= 0:
            return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        self.value_counts = file.value_counts()
        self.null_counts = file.null_value_counts()
        self.lower_bounds = file.lower_bounds()
        self.upper_bounds = file.upper_bounds()

        return ExpressionVisitors.visit(self.expr, self)

    def always_true(self):
        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def always_false(self):
        return MetricsEvalVisitor.ROWS_CANNOT_MATCH

    def not_(self, result):
        return not result

    def and_(self, left_result, right_result):
        return left_result and right_result

    def or_(self, left_result, right_result):
        return left_result or right_result

    def is_null(self, ref):
        id = ref.field.field_id

        if self.struct.field(id=id) is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.null_counts is not None and self.null_counts.get(id, -1) == 0:
            return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def not_null(self, ref):
        id = ref.field.field_id

        if self.struct.field(id=id) is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.value_counts is not None and id in self.value_counts and id in self.null_counts \
                and self.value_counts.get(id) - self.null_counts.get(id) == 0:
            return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def lt(self, ref, lit):
        id = ref.field.field_id
        field = self.struct.field(id=id)

        if field is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.lower_bounds is not None and id in self.lower_bounds:
            lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
            if lower >= lit.value:
                return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def lt_eq(self, ref, lit):
        id = ref.field.field_id
        field = self.struct.field(id=id)

        if field is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.lower_bounds is not None and id in self.lower_bounds:
            lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
            if lower > lit.value:
                return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def gt(self, ref, lit):
        id = ref.field.field_id
        field = self.struct.field(id=id)

        if field is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.upper_bounds is not None and id in self.upper_bounds:
            upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))
            if upper <= lit.value:
                return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def gt_eq(self, ref, lit):
        id = ref.field.field_id
        field = self.struct.field(id=id)

        if field is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.upper_bounds is not None and id in self.upper_bounds:
            upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))
            if upper < lit.value:
                return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def eq(self, ref, lit):
        id = ref.field.field_id
        field = self.struct.field(id=id)

        if field is None:
            raise RuntimeError("Cannot filter by nested column: %s" % self.schema.find_field(id))

        if self.lower_bounds is not None and id in self.lower_bounds:
            lower = Conversions.from_byte_buffer(field.type, self.lower_bounds.get(id))
            if lower > lit.value:
                return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        if self.upper_bounds is not None and id in self.upper_bounds:
            upper = Conversions.from_byte_buffer(field.type, self.upper_bounds.get(id))
            if upper < lit.value:
                return MetricsEvalVisitor.ROWS_CANNOT_MATCH

        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def not_eq(self, ref, lit):
        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def in_(self, ref, lit):
        return MetricsEvalVisitor.ROWS_MIGHT_MATCH

    def not_in(self, ref, lit):
        return MetricsEvalVisitor.ROWS_MIGHT_MATCH



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


__all__ = ["AtomicInteger",
           "PackingIterator",
           "PLANNER_THREAD_POOL_SIZE_PROP",
           "SCAN_THREAD_POOL_ENABLED",
           "str_as_bool",
           "WORKER_THREAD_POOL_SIZE_PROP",
           ]

from .atomic_integer import AtomicInteger
from .bin_packing import PackingIterator

PLANNER_THREAD_POOL_SIZE_PROP = "iceberg.planner.num-threads"
WORKER_THREAD_POOL_SIZE_PROP = "iceberg.worker.num-threads"
SCAN_THREAD_POOL_ENABLED = "iceberg.scan.plan-in-worker-pool"


def str_as_bool(str_var):
    return str_var is not None and str_var.lower() == "true"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import math
import struct
import sys

import mmh3

from .transform import Transform
from .transform_util import TransformUtil
from ..expressions import (Expressions,
                           JAVA_MAX_INT,
                           Operation)
from ..types.types import (IntegerType,
                           TypeID)
from ...api.types.conversions import Conversions


class Bucket(Transform):
    MURMUR3 = mmh3

    BUCKET_TYPE = {TypeID.DATE: lambda n: BucketInteger(n),
                   TypeID.INTEGER: lambda n: BucketInteger(n),
                   TypeID.TIME: lambda n: BucketLong(n),
                   TypeID.TIMESTAMP: lambda n: BucketLong(n),
                   TypeID.LONG: lambda n: BucketLong(n),
                   TypeID.DECIMAL: lambda n: BucketDecimal(n),
                   TypeID.STRING: lambda n: BucketString(n),
                   TypeID.FIXED: lambda n: BucketByteBuffer(n),
                   TypeID.BINARY: lambda n: BucketByteBuffer(n),
                   TypeID.UUID: lambda n: BucketUUID(n)}

    @staticmethod
    def get(type_var, n):
        bucket_type_func = Bucket.BUCKET_TYPE.get(type_var.type_id)
        if not bucket_type_func:
            raise RuntimeError("Cannot bucket by type: %s" % type_var)
        return bucket_type_func(n)

    def __init__(self, n):
        self.n = n

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        if other is None or not isinstance(other, Bucket):
            return False

        return self.n == other.n

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return Bucket.__class__, self.n

    def __repr__(self):
        return "Bucket[%s]" % self.n

    def __str__(self):
        return "bucket[%s]" % self.n

    def apply(self, value):
        return (self.hash(value) & JAVA_MAX_INT) % self.n

    def hash(self):
        raise NotImplementedError()

    def project(self, name, predicate):
        if predicate.op == Operation.EQ:
            return Expressions.predicate(predicate.op, name, self.apply(predicate.lit.value))

    def project_strict(self, name, predicate):
        if predicate.op == Operation.NOT_EQ:
            return Expressions.predicate(predicate.op, name, self.apply(predicate.lit.value))

    def get_result_type(self, source_type):
        return IntegerType.get()


class BucketInteger(Bucket):

    def __init__(self, n):
        super(BucketInteger, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(struct.pack("q", value))

    def can_transform(self, type_var):
        return type_var.type_id in [TypeID.INTEGER, TypeID.DATE]


class BucketLong(Bucket):
    def __init__(self, n):
        super(BucketLong, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(struct.pack("q", value))

    def can_transform(self, type_var):
        return type_var.type_id in [TypeID.LONG,
                                    TypeID.TIME,
                                    TypeID.TIMESTAMP]


class BucketFloat(Bucket):
    def __init__(self, n):
        super(BucketFloat, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(struct.pack("d", value))

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.FLOAT


class BucketDouble(Bucket):
    def __init__(self, n):
        super(BucketDouble, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(struct.pack("d", value))

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.DOUBLE


class BucketDecimal(Bucket):

    def __init__(self, n):
        super(BucketDecimal, self).__init__(n)

    def hash(self, value):
        # to-do: unwrap to_bytes func since python2 support is being removed
        unscaled_value = TransformUtil.unscale_decimal(value)
        number_of_bytes = int(math.ceil(unscaled_value.bit_length() / 8))
        return Bucket.MURMUR3.hash(to_bytes(unscaled_value, number_of_bytes, byteorder='big'))

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.DECIMAL


class BucketString(Bucket):
    def __init__(self, n):
        super(BucketString, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(value)

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.STRING


class BucketByteBuffer(Bucket):
    def __init__(self, n):
        super(BucketByteBuffer, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(value)

    def can_transform(self, type_var):
        return type_var.type_id in [TypeID.BINARY, TypeID.FIXED]


class BucketUUID(Bucket):
    def __init__(self, n):
        super(BucketUUID, self).__init__(n)

    def hash(self, value):
        return Bucket.MURMUR3.hash(Conversions.to_byte_buffer(TypeID.UUID, value))

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.UUID


def to_bytes(n, length, byteorder='big'):
    if sys.version_info >= (3, 0):
        return n.to_bytes(length, byteorder=byteorder)
    h = '%x' % n
    s = ('0' * (len(h) % 2) + h).zfill(length * 2).decode('hex')
    return s if byteorder == 'big' else s[::-1]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List

from iceberg.api import Schema
from iceberg.api.types import get_projected_ids


def prune_columns(file_schema: Schema, expected_schema: Schema) -> List[str]:
    """
    Given two Iceberg schema's returns a list of column_names for all id's in the
    file schema that are projected in the expected schema

    Parameters
    ----------
    file_schema : iceberg.api.Schema
        An Iceberg schema of the file being read
    expected_schema : iceberg.api.Schema
        An Iceberg schema of the final projection
    Returns
    -------
    list
        The column names in the file that matched ids in the expected schema
    """
    return [column.name for column in file_schema.as_struct().fields
            if column.id in get_projected_ids(expected_schema)]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging

from .expression import (And,
                         FALSE,
                         Not,
                         Operation,
                         Or,
                         TRUE)
from .expression_parser import parse_expr_string
from .predicate import (Predicate,
                        UnboundPredicate)
from .reference import NamedReference

_logger = logging.getLogger(__name__)


class Expressions(object):

    @staticmethod
    def and_(left, right):
        if left == Expressions.always_false() or right == Expressions.always_false():
            return Expressions.always_false()
        elif left == Expressions.always_true():
            return right
        elif right == Expressions.always_true():
            return left

        return And(left, right)

    @staticmethod
    def or_(left, right):
        if left == Expressions.always_true() or right == Expressions.always_true():
            return Expressions.always_true()
        elif left == Expressions.always_false():
            return right
        elif right == Expressions.always_false():
            return left

        return Or(left, right)

    @staticmethod
    def not_(child):
        if child == Expressions.always_true():
            return Expressions.always_false()
        elif child == Expressions.always_false():
            return Expressions.always_true()
        elif isinstance(child, Not):
            return child.child

        return Not(child)

    @staticmethod
    def is_null(name):
        return UnboundPredicate(Operation.IS_NULL, Expressions.ref(name))

    @staticmethod
    def not_null(name):
        return UnboundPredicate(Operation.NOT_NULL, Expressions.ref(name))

    @staticmethod
    def is_nan(name):
        return UnboundPredicate(Operation.IS_NAN, Expressions.ref(name))

    @staticmethod
    def not_nan(name):
        return UnboundPredicate(Operation.NOT_NAN, Expressions.ref(name))

    @staticmethod
    def less_than(name, value):
        return UnboundPredicate(Operation.LT, Expressions.ref(name), value)

    @staticmethod
    def less_than_or_equal(name, value):
        return UnboundPredicate(Operation.LT_EQ, Expressions.ref(name), value)

    @staticmethod
    def greater_than(name, value):
        return UnboundPredicate(Operation.GT, Expressions.ref(name), value)

    @staticmethod
    def greater_than_or_equal(name, value):
        return UnboundPredicate(Operation.GT_EQ, Expressions.ref(name), value)

    @staticmethod
    def equal(name, value):
        return UnboundPredicate(Operation.EQ, Expressions.ref(name), value)

    @staticmethod
    def not_equal(name, value):
        return UnboundPredicate(Operation.NOT_EQ, Expressions.ref(name), value)

    @staticmethod
    def predicate(op, name, value=None, lit=None):
        if value is not None and op not in (Operation.IS_NULL, Operation.NOT_NULL):
            return UnboundPredicate(op, Expressions.ref(name), value)
        elif lit is not None and op not in (Operation.IS_NULL, Operation.NOT_NULL):
            return UnboundPredicate(op, Expressions.ref(name), value)
        elif op in (Operation.IS_NULL, Operation.NOT_NULL):
            if value is not None or lit is not None:
                raise RuntimeError("Cannot create {} predicate inclusive a value".format(op))
            return UnboundPredicate(op, Expressions.ref(name))
        else:
            raise RuntimeError("Cannot create {} predicate without a value".format(op))

    @staticmethod
    def always_true():
        return TRUE

    @staticmethod
    def always_false():
        return FALSE

    @staticmethod
    def rewrite_not(expr):
        return ExpressionVisitors.visit(expr, RewriteNot.get()) # noqa

    @staticmethod
    def ref(name):
        return NamedReference(name)

    @staticmethod
    def convert_string_to_expr(predicate_string):
        expr_map = {"and": (Expressions.and_,),
                    "eq": (Expressions.equal,),
                    "exists": (Expressions.not_null,),
                    "gt": (Expressions.greater_than,),
                    "gte": (Expressions.greater_than_or_equal,),
                    "lt": (Expressions.less_than,),
                    "lte": (Expressions.less_than_or_equal,),
                    "missing": (Expressions.is_null,),
                    "neq": (Expressions.not_equal,),
                    "not": (Expressions.not_,),
                    "or": (Expressions.or_,)}

        return parse_expr_string(predicate_string, expr_map)


class ExpressionVisitors(object):

    @staticmethod
    def visit(expr, visitor):
        if isinstance(expr, Predicate):
            return visitor.predicate(expr)

        if expr.op() == Operation.TRUE:
            return visitor.always_true()
        elif expr.op() == Operation.FALSE:
            return visitor.always_false()
        elif expr.op() == Operation.NOT:
            return visitor.not_(ExpressionVisitors.visit(expr.child, visitor))
        elif expr.op() == Operation.AND:
            return visitor.and_(ExpressionVisitors.visit(expr.left, visitor),
                                ExpressionVisitors.visit(expr.right, visitor))
        elif expr.op() == Operation.OR:
            return visitor.or_(ExpressionVisitors.visit(expr.left, visitor),
                               ExpressionVisitors.visit(expr.right, visitor))
        else:
            raise RuntimeError("Unknown operation: {}".format(expr.op()))

    class ExpressionVisitor(object):

        def always_true(self):
            return NotImplementedError()

        def always_false(self):
            return NotImplementedError()

        def not_(self, result):
            return NotImplementedError()

        def and_(self, left_result, right_result):
            return NotImplementedError()

        def or_(self, left_result, right_result):
            return NotImplementedError()

        def predicate(self, pred):
            return NotImplementedError()

    class BoundExpressionVisitor(ExpressionVisitor):

        def __init__(self):
            super(ExpressionVisitors.BoundExpressionVisitor, self).__init__()

        def is_null(self, ref):
            return NotImplementedError()

        def not_null(self, ref):
            return NotImplementedError()

        def is_nan(self, ref):
            return NotImplementedError()

        def not_nan(self, ref):
            return NotImplementedError()

        def lt(self, ref, lit):
            return NotImplementedError()

        def lt_eq(self, ref, lit):
            return NotImplementedError()

        def gt(self, ref, lit):
            return NotImplementedError()

        def gt_eq(self, ref, lit):
            return None

        def eq(self, ref, lit):
            return None

        def not_eq(self, ref, lit):
            return None

        def in_(self, ref, lit):
            return None

        def not_in(self, ref, lit):
            return None

        def predicate(self, pred): # noqa

            if isinstance(pred, UnboundPredicate):
                raise RuntimeError("Not a bound Predicate: {}".format(pred))

            if pred.op == Operation.IS_NULL:
                return self.is_null(pred.ref)
            elif pred.op == Operation.NOT_NULL:
                return self.not_null(pred.ref)
            elif pred.op in [Operation.IS_NAN, Operation.NOT_NAN]:
                raise NotImplementedError("IS_NAN and NOT_NAN not fully implemented for expressions")
            elif pred.op == Operation.LT:
                return self.lt(pred.ref, pred.lit)
            elif pred.op == Operation.LT_EQ:
                return self.lt_eq(pred.ref, pred.lit)
            elif pred.op == Operation.GT:
                return self.gt(pred.ref, pred.lit)
            elif pred.op == Operation.GT_EQ:
                return self.gt_eq(pred.ref, pred.lit)
            elif pred.op == Operation.EQ:
                return self.eq(pred.ref, pred.lit)
            elif pred.op == Operation.NOT_EQ:
                return self.not_eq(pred.ref, pred.lit)
            elif pred.op == Operation.IN:
                return self.in_(pred.ref, pred.lit)
            elif pred.op == Operation.NOT_IN:
                return self.not_in(pred.ref, pred.lit)
            else:
                raise RuntimeError("Unknown operation for Predicate: {}".format(pred.op))


class RewriteNot(ExpressionVisitors.ExpressionVisitor):
    __instance = None

    @staticmethod
    def get():
        if RewriteNot.__instance is None:
            RewriteNot()
        return RewriteNot.__instance

    def __init__(self):
        if RewriteNot.__instance is not None:
            raise Exception("Multiple RewriteNot Types created")
        RewriteNot.__instance = self

    def always_true(self):
        return Expressions.always_true()

    def always_false(self):
        return Expressions.always_false()

    def not_(self, result):
        return result.negate()

    def and_(self, left_result, right_result):
        return Expressions.and_(left_result, right_result)

    def or_(self, left_result, right_result):
        return Expressions.or_(left_result, right_result)

    def predicate(self, pred):
        return pred



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import decimal

from iceberg.api.expressions import Expressions, Operation


class ProjectionUtil(object):
    @staticmethod
    def truncate_integer(name, pred, transform):
        boundary = pred.lit.value
        if pred.op == Operation.LT:
            return Expressions.predicate(Operation.LT_EQ, name, transform.apply(boundary - 1))
        elif pred.op == Operation.LT_EQ:
            return Expressions.predicate(Operation.LT_EQ, name, transform.apply(boundary))
        elif pred.op == Operation.GT:
            return Expressions.predicate(Operation.GT_EQ, name, transform.apply(boundary + 1))
        elif pred.op == Operation.GT_EQ:
            return Expressions.predicate(Operation.GT_EQ, name, transform.apply(boundary))
        elif pred.op == Operation.EQ:
            return Expressions.predicate(pred.op, name, transform.apply(boundary))

    def truncate_long(name, pred, transform):
        return ProjectionUtil.truncate_integer(name, pred, transform)

    def truncate_decimal(name, pred, transform):
        boundary = pred.lit.value

        if pred.op == Operation.LT:
            minus_one = boundary - decimal.Decimal(1)
            return Expressions.predicate(Operation.LT_EQ, name, transform.apply(minus_one))
        elif pred.op == Operation.LT_EQ:
            return Expressions.predicate(Operation.LT_EQ, name, transform.apply(boundary))
        elif pred.op == Operation.GT:
            plus_one = boundary + decimal.Decimal(1)
            return Expressions.predicate(Operation.GT_EQ, name, transform.apply(plus_one))
        elif pred.op == Operation.GT_EQ:
            return Expressions.predicate(Operation.GT_EQ, name, transform.apply(boundary))
        elif pred.op == Operation.EQ:
            return Expressions.predicate(pred.op, name, transform.apply(boundary))

    def truncate_array(name, pred, transform):
        boundary = pred.lit.value

        if pred.op == Operation.LT or pred.op == Operation.LT_EQ:
            return Expressions.predicate(Operation.LT_EQ, name, transform.apply(boundary))
        elif pred.op == Operation.GT or pred.op == Operation.GT_EQ:
            return Expressions.predicate(Operation.GT_EQ, name, transform.apply(boundary))
        elif pred.op == Operation.EQ:
            return Expressions.predicate(pred.op, name, transform.apply(boundary))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json

from iceberg.api import Schema
from iceberg.api.types import (from_primitive_string,
                               ListType,
                               MapType,
                               NestedField,
                               StructType,
                               Type,
                               TypeID)


class SchemaParser(object):

    TYPE = "type"
    STRUCT = "struct"
    LIST = "list"
    MAP = "map"
    FIELDS = "fields"
    ELEMENT = "element"
    KEY = "key"
    VALUE = "value"
    DOC = "doc"
    NAME = "name"
    ID = "id"
    ELEMENT_ID = "element-id"
    KEY_ID = "key-id"
    VALUE_ID = "value-id"
    REQUIRED = "required"
    ELEMENT_REQUIRED = "element-required"
    VALUE_REQUIRED = "value-required"

    @staticmethod
    def to_json(type_var, indent=None):
        return json.dumps(SchemaParser.to_dict(type_var), indent=indent)

    @staticmethod
    def to_dict(type_var):
        if isinstance(type_var, Schema):
            struct = type_var.as_struct()
        elif isinstance(type_var, Type):
            struct = type_var

        return SchemaParser._type_to_dict(struct)

    @staticmethod
    def _type_to_dict(type_var):
        if type_var.is_primitive_type():
            return SchemaParser._primitive_to_dict(type_var)
        else:
            nested = type_var.as_nested_type()
            if type_var.type_id == TypeID.STRUCT:
                return SchemaParser._struct_to_dict(nested)
            elif type_var.type_id == TypeID.LIST:
                return SchemaParser._list_to_dict(nested)
            elif type_var.type_id == TypeID.MAP:
                return SchemaParser._map_to_dict(nested)
            else:
                raise RuntimeError("Cannot write unknown type: %s" % type)

    @staticmethod
    def _primitive_to_dict(type_var):
        return str(type_var)

    @staticmethod
    def _struct_to_dict(struct):
        return {SchemaParser.TYPE: SchemaParser.STRUCT,
                SchemaParser.FIELDS: [SchemaParser._struct_field_to_dict(field) for field in struct.fields]}

    @staticmethod
    def _struct_field_to_dict(field):
        schema = {SchemaParser.ID: field.field_id,
                  SchemaParser.NAME: field.name,
                  SchemaParser.REQUIRED: field.is_required,
                  SchemaParser.TYPE: SchemaParser._type_to_dict(field.type)}
        if field.doc is not None:
            schema[SchemaParser.DOC] = field.doc

        return schema

    @staticmethod
    def _list_to_dict(list_type):
        return {SchemaParser.TYPE: SchemaParser.LIST,
                SchemaParser.ELEMENT_ID: list_type.element_id,
                SchemaParser.ELEMENT: SchemaParser._type_to_dict(list_type.element_type),
                SchemaParser.ELEMENT_REQUIRED: list_type.is_element_required()}

    @staticmethod
    def _map_to_dict(map_type):
        return {SchemaParser.TYPE: SchemaParser.MAP,
                SchemaParser.KEY_ID: map_type.key_id(),
                SchemaParser.KEY: SchemaParser._type_to_dict(map_type.key_type()),
                SchemaParser.VALUE_ID: map_type.value_id(),
                SchemaParser.VALUE: SchemaParser._type_to_dict(map_type.value_type()),
                SchemaParser.VALUE_REQUIRED: map_type.is_value_required()}

    @staticmethod
    def type_from_dict(dict_obj):
        if isinstance(dict_obj, (str, bool, int, float)):
            return from_primitive_string(dict_obj)
        else:
            type_str = dict_obj.get(SchemaParser.TYPE)
            if type_str.upper() == "STRUCT":
                return SchemaParser.struct_from_dict(dict_obj)
            elif type_str.upper() == "LIST":
                return SchemaParser.list_from_dict(dict_obj)
            elif type_str.upper() == "MAP":
                return SchemaParser.map_from_dict(dict_obj)
            else:
                raise RuntimeError("Cannot parse type from dict: %s" % dict_obj)

    @staticmethod
    def struct_from_dict(dict_obj):
        struct_fields = list()
        fields = dict_obj.get(SchemaParser.FIELDS)
        for field in fields:
            field_id = field.get(SchemaParser.ID)
            field_name = field.get(SchemaParser.NAME)
            field_type = SchemaParser.type_from_dict(field.get(SchemaParser.TYPE))
            field_doc = field.get(SchemaParser.DOC)

            if field.get(SchemaParser.REQUIRED):
                struct_fields.append(NestedField.required(field_id, field_name, field_type, field_doc))
            else:
                struct_fields.append(NestedField.optional(field_id, field_name, field_type, field_doc))

        return StructType.of(struct_fields)

    @staticmethod
    def list_from_dict(dict_obj):
        elem_id = dict_obj.get(SchemaParser.ELEMENT_ID)
        elem_type = SchemaParser.type_from_dict(dict_obj.get(SchemaParser.ELEMENT))
        is_required = dict_obj.get(SchemaParser.REQUIRED)

        if is_required:
            return ListType.of_required(elem_id, elem_type)
        else:
            return ListType.of_optional(elem_id, elem_type)

    @staticmethod
    def map_from_dict(dict_obj):
        key_id = dict_obj.get(SchemaParser.KEY_ID)
        key_type = SchemaParser.type_from_dict(dict_obj.get(SchemaParser.KEY))

        value_id = dict_obj.get(SchemaParser.VALUE_ID)
        value_type = SchemaParser.type_from_dict(dict_obj.get(SchemaParser.VALUE))

        is_required = dict_obj.get(SchemaParser.VALUE_REQUIRED)

        if is_required:
            return MapType.of_required(key_id, value_id, key_type, value_type)
        else:
            return MapType.of_optional(key_id, value_id, key_type, value_type)

    @staticmethod
    def from_json(json_obj):
        if isinstance(json_obj, str):
            type_var = SchemaParser.type_from_dict(json.loads(json_obj))
        else:
            type_var = SchemaParser.type_from_dict(json_obj)

        if type_var is not None and type_var.is_nested_type() and type_var.as_nested_type().is_struct_type():
            return Schema(type_var.as_nested_type().as_struct_type().fields)
        else:
            raise RuntimeError("Cannot create schema, not a struct type: %s" % type_var)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import logging

from iceberg.api import Schema
from iceberg.api.types import (BinaryType,
                               BooleanType,
                               DateType,
                               DecimalType,
                               DoubleType,
                               FixedType,
                               FloatType,
                               IntegerType,
                               ListType,
                               LongType,
                               MapType,
                               NestedField,
                               StringType,
                               StructType,
                               TimestampType)
from iceberg.api.types import Type
import pyarrow as pa
from pyarrow.parquet import lib, ParquetFile

_logger = logging.getLogger(__name__)

arrow_type_map = {lib.Type_BOOL: lambda x=None: BooleanType.get(),
                  lib.Type_DATE32: lambda x=None: DateType.get(),
                  lib.Type_DECIMAL128: lambda x=None: DecimalType.of(x.precision, x.scale),
                  lib.Type_DOUBLE: lambda x=None: DoubleType.get(),
                  lib.Type_FIXED_SIZE_BINARY: lambda x=None: FixedType.of_length(x.byte_width),
                  lib.Type_BINARY: lambda x=None: BinaryType.get(),
                  lib.Type_FLOAT: lambda x=None: FloatType.get(),
                  lib.Type_STRING: lambda x=None: StringType.get(),
                  lib.Type_INT32: lambda x=None: IntegerType.get(),
                  lib.Type_INT64: lambda x=None: LongType.get(),
                  lib.Type_TIMESTAMP: lambda x=None: (TimestampType.without_timezone()
                                                      if x.tz is None
                                                      else TimestampType.with_timezone())
                  }


def get_nested_field(field_id: int, field_name: str, field_type: Type, nullable: bool) -> NestedField:
    if nullable:
        return NestedField.optional(field_id, field_name, field_type)
    else:
        return NestedField.required(field_id, field_name, field_type)


def get_list_field(col_type: pa.ListType) -> ListType:
    if col_type.value_field.nullable:
        return ListType.of_optional(int(col_type.value_field.metadata[b"PARQUET:field_id"].decode("utf-8")),
                                    arrow_type_map[col_type.value_field.type.id](col_type.value_field.type))
    else:
        return ListType.of_required(col_type.value_field.metadata[b"PARQUET:field_id"].decode("utf-8"),
                                    arrow_type_map[col_type.value_field.type.id](col_type.value_field.type))


def get_field(col: pa.Field) -> NestedField:
    try:
        return get_nested_field(int(col.metadata[b"PARQUET:field_id"].decode("utf-8")),
                                col.name,
                                arrow_type_map[col.type.id](col.type),
                                col.nullable)
    except KeyError:
        _logger.error(f"\t{int(col.metadata[b'PARQUET:field_id'].decode('utf-8'))}: {col.type}")
        raise


def get_struct_field(col_type: pa.StructType) -> StructType:
    if col_type.num_fields > 0:
        if col_type[0].name == "map":
            return get_inferred_map(col_type)
        else:
            return StructType.of([get_field(child) for child in col_type])
    else:
        raise RuntimeError("Unable to convert type to iceberg %s" % col_type)


def get_inferred_map(col_type: pa.StructType):
    base_field = col_type[0]
    key_field = get_field(col_type[0].type.value_field.type[0])
    value_field = get_field(col_type[0].type.value_field.type[1])

    if base_field.nullable:
        return MapType.of_optional(key_field.field_id, value_field.field_id,
                                   key_field.type, value_field.type)
    else:
        return MapType.of_required(key_field.field_id, value_field.field_id,
                                   key_field.type, value_field.type)


def get_map_field(col):
    # Spark Map type gets serialized to Struct<List<Struct<>>>
    raise NotImplementedError("Arrow Map type not implemented yet")


# adding function mappings for nested types
arrow_type_map.update({lib.Type_LIST: get_list_field,
                       lib.Type_MAP: get_map_field,
                       lib.Type_STRUCT: get_struct_field})


def convert_parquet_to_iceberg(parquet_file: ParquetFile) -> Schema:  # noqa: ignore=C901
    """
    Given two Iceberg schema's returns a list of column_names for all id's in the
    file schema that are projected in the expected schema

    Parameters
    ----------
    parquet_file : pyarrow.parquet.ParquetFile
        The Parquet File to use to extract the iceberg schema

    Returns
    -------
    iceberg.api.Schema
        returns an equivalent iceberg Schema based on the arrow schema read from the file
    """
    return arrow_to_iceberg(parquet_file.schema_arrow)


def arrow_to_iceberg(arrow_schema: pa.Schema) -> Schema:
    """
    Use an arrow schema, which contains the field_id metadata, to create an equivalent iceberg Schema

    Parameters
    ----------
    arrow_schema : pyarrow.Schema
        An Arrow schema with the parquet field_id metadata

    Returns
    -------
    iceberg.api.Schema
        returns an equivalent iceberg Schema based on the arrow schema read from the file
    """
    return Schema([get_field(col) for col in arrow_schema])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .types import StructType
from .types import type_util
"""
  The schema of a data table.

"""


class Schema(object):
    NEWLINE = '\n'
    ALL_COLUMNS = "*"

    def __init__(self, *argv):
        aliases = None
        if len(argv) == 1 and isinstance(argv[0], (list, tuple)):
            columns = argv[0]
        elif len(argv) == 2 and isinstance(argv[0], list) and isinstance(argv[1], dict):
            columns = argv[0]
            aliases = argv[1]
        else:
            columns = argv

        self.struct = StructType.of(columns)
        self._alias_to_id = None
        self._id_to_alias = None
        if aliases is not None:
            self._alias_to_id = dict(aliases)
            self._id_to_alias = {v: k for k, v in self._alias_to_id.items()}

        self._id_to_field = None
        self._name_to_id = type_util.index_by_name(self.struct)
        self._id_to_name = {v: k for k, v in self._name_to_id.items()}
        self._lowercase_name_to_id = {k.lower(): v for k, v in self._name_to_id.items()}

    def as_struct(self):
        return self.struct

    def get_aliases(self):
        return self._alias_to_id

    def lazy_id_to_field(self):
        from .types import index_by_id
        if self._id_to_field is None:
            self._id_to_field = index_by_id(self.struct) # noqa

        return self._id_to_field

    def lazy_name_to_id(self):
        from .types import index_by_name
        if self._name_to_id is None:
            self._name_to_id = index_by_name(self.struct)
            self._id_to_name = {v: k for k, v in self._name_to_id.items()}
            self._lowercase_name_to_id = {k.lower(): v for k, v in self._name_to_id.items()}

        return self._name_to_id

    def lazy_lowercase_name_to_id(self):
        from .types import index_by_name
        if self._lowercase_name_to_id is None:
            self._name_to_id = index_by_name(self.struct)
            self._id_to_name = {v: k for k, v in self._name_to_id.items()}
            self._lowercase_name_to_id = {k.lower(): v for k, v in self._name_to_id.items()}

        return self._lowercase_name_to_id

    def columns(self):
        return self.struct.fields

    def find_type(self, name):
        if isinstance(name, int):
            field = self.lazy_id_to_field().get(name)
            if field:
                return field.type
        elif isinstance(name, str):
            id = self.lazy_name_to_id().get(name)
            if id:
                return self.find_type(id)
        else:
            raise RuntimeError("Invalid Column (could not find): %s" % name)

    def find_field(self, id):
        if isinstance(id, int):
            return self.lazy_id_to_field().get(id)

        if not id:
            raise ValueError("Invalid Column Name (empty)")

        id = self.lazy_name_to_id().get(id)
        if id is not None:
            return self.lazy_id_to_field().get(id)

    def case_insensitive_find_field(self, name):
        if name is None:
            raise ValueError("Invalid Column Name (empty)")

        id = self.lazy_lowercase_name_to_id().get(name.lower())
        if id is not None:
            return self.lazy_id_to_field().get(id)

        return None

    def find_column_name(self, id):
        if isinstance(id, int):
            field = self.lazy_id_to_field().get(id)

            return None if field is None else field.name

    def alias_to_id(self, alias):
        if self._alias_to_id:
            return self._alias_to_id.get(alias)

    def id_to_alias(self, field_id):
        if self._id_to_alias:
            return self._id_to_alias.get(field_id)

    def select(self, cols):
        return self._internal_select(True, cols)

    def case_insensitive_select(self, cols):
        return self._internal_select(False, cols)

    def _internal_select(self, case_sensitive, cols):
        from .types import select

        if Schema.ALL_COLUMNS in cols:
            return self

        selected = set()
        for name in cols:
            if case_sensitive:
                field_id = self.lazy_name_to_id().get(name)
            else:
                field_id = self.lazy_lowercase_name_to_id().get(name.lower())

            if field_id is not None:
                selected.add(field_id)

        return select(self, selected)

    def __len__(self):
        return len(self.struct.fields)

    def __repr__(self):
        return "Schema(%s)" % ",".join([str(field) for field in self.struct.fields])

    def __str__(self):
        return "table {\n%s\n}" % Schema.NEWLINE.join([" " + str(field) for field in self.struct.fields])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from copy import deepcopy
import functools

from iceberg.api import DataOperations
from iceberg.api.expressions import Expressions, Literal, Operation, UnboundPredicate
from iceberg.api.types import TimestampType

from .manifest_group import ManifestGroup
from .util import str_as_bool

TIMESTAMP_RANGE_MAP = {Operation.LT: lambda min_val, max_val, val: (min_val, val - 1 if val - 1 < max_val else max_val),
                       Operation.LT_EQ: lambda min_val, max_val, val: (min_val, val if val < max_val else max_val),
                       Operation.GT: lambda min_val, max_val, val: (val - 1 if val - 1 > min_val else min_val, max_val),
                       Operation.GT_EQ: lambda min_val, max_val, val: (val if val > min_val else min_val, max_val),
                       Operation.EQ: lambda min_val, max_val, val: (val if val > min_val else min_val,
                                                                    val if val < max_val else max_val)}


class ScanSummary(object):

    IGNORED_OPERATIONS = {DataOperations.DELETE, DataOperations.REPLACE}
    SCAN_SUMMARY_COLUMNS = ["partition", "record_count", "file_size_in_bytes"]


class ScanSummaryBuilder(object):

    TIMESTAMP_NAMES = {"dateCreated", "lastUpdated"}

    def __init__(self, scan):
        self.scan = scan
        self.table = scan.table
        self.ops = self.table.ops
        self.snapshot_timestamps = {snap.snapshot_id: snap.timestamp_millis
                                    for snap in self.table.snapshots()}
        self._throw_if_limited = False
        self.force_use_manifests = False
        self._limit = 2**31 - 1

        self.time_filters = list()

    def add_timestamp_filter(self, filter):
        self.throw_if_limited()
        self.time_filters.append(filter)
        return self

    def after(self, timestamp):
        self.add_timestamp_expression(timestamp, Expressions.greater_than_or_equal)
        return self

    def before(self, timestamp):
        self.add_timestamp_expression(timestamp, Expressions.less_than_or_equal)
        return self

    def add_timestamp_expression(self, timestamp, expr_func):
        if isinstance(timestamp, str):
            timestamp = Literal.of(timestamp).to(TimestampType.without_timezone()).value / 1000

        self.add_timestamp_filter(expr_func("timestamp_ms", timestamp))
        return self

    def throw_if_limited(self):
        self._throw_if_limited = True
        return self

    def limit(self, num_partitions):
        self._limit = num_partitions
        return self

    def use_manifests(self):
        self.force_use_manifests = True
        return self

    def remove_time_filters(self, expressions, expression):
        if expression.op == Operation.AND:
            self.remove_time_filters(expressions, expression.left)
            self.remove_time_filters(expressions, expression.right)
            return
        elif isinstance(expression, UnboundPredicate):
            pred = expression
            ref = pred.ref
            lit = pred.lit

            if ref.name in ScanSummaryBuilder.TIMESTAMP_NAMES:
                ts_literal = lit.to(TimestampType.without_timezone())
                millis = ScanSummaryBuilder.to_millis(ts_literal.value)
                self.add_timestamp_filter(Expressions.predicate(pred.op, "timestamp_ms", millis))
                return

        expressions.append(expression)

    def build(self):
        if self.table.current_snapshot() is None:
            return dict()

        filters = list()
        self.remove_time_filters(filters, Expressions.rewrite_not(self.scan.row_filter))
        row_filter = self.join_filters(filters)

        if len(self.time_filters) == 0:
            return self.from_manifest_scan(self.table.current_snapshot().manifests, row_filter)

        min_timestamp, max_timestamp = self.timestamp_range(self.time_filters)
        oldest_snapshot = self.table.current_snapshot()
        for key, val in self.snapshot_timestamps.items():
            if val < oldest_snapshot.timestamp_millis:
                oldest_snapshot = self.ops.current().snapshot(key)

        # if oldest known snapshot is in the range, then there may be an expired snapshot that has
        # been removed that matched the range. because the timestamp of that snapshot is unknown,
        # it can't be included in the results and the results are not reliable."""
        if oldest_snapshot.timestamp_millis >= min_timestamp and oldest_snapshot <= max_timestamp:
            raise RuntimeError("Cannot satisfy time filters: time range may include expired snapshots")

        snapshots = [snapshot for snapshot in ScanSummaryBuilder.snapshots_in_time_range(self.ops.current(),
                                                                                         min_timestamp,
                                                                                         max_timestamp)
                     if snapshot.operation not in ScanSummary.IGNORED_OPERATIONS]

        result = self.from_partition_summaries(snapshots)
        if result is not None and not self.force_use_manifests:
            return result

        # filter down to the the set of manifest files that were created in the time range, ignoring
        # the snapshots created by delete or replace operations. this is complete because it finds
        # files in the snapshot where they were added to the dataset in either an append or an
        # overwrite. if those files are later compacted with a replace or deleted, those changes are
        # ignored.
        manifests_to_scan = list()
        snapshot_ids = set()

        for snap in snapshots:
            snapshot_ids.add(snap)
            for manifest in snap.manifests:
                if manifest.snapshot_id is not None or manifest.snapshot_id == snap.snapshot_id:
                    manifests_to_scan.append(manifest)

        return self.from_manifest_scan(manifests_to_scan, row_filter, True)

    def from_manifest_scan(self, manifests, row_filter, ignore_existing=False):
        top_n = TopN(self._limit, self._throw_if_limited, lambda x, y: 0 if x == y else -1 if x < y else 1)

        entries = (ManifestGroup(self.ops, manifests)
                   .filter_data(row_filter)
                   .ignore_deleted()
                   .ignore_existing(ignore_existing)
                   .select(ScanSummary.SCAN_SUMMARY_COLUMNS)
                   .entries())

        spec = self.table.spec()
        for entry in entries:
            timestamp = self.snapshot_timestamps.get(entry.snapshot_id)
            partition = spec.partition_to_path(entry.file.partition())
            top_n.update(partition,
                         lambda metrics: ((metrics if metrics is not None else PartitionMetrics())
                                          .update_from_file(entry.file, timestamp)))

        return top_n.get()

    def from_partition_summaries(self, snapshots):
        # try to build the result from snapshot metadata, but fall back if:
        # any snapshot has no summary
        # any snapshot has
        top_n = TopN(self._limit, self._throw_if_limited, lambda x, y: 0 if x == y else -1 if x < y else 1)

        for snap in snapshots:
            if snap.operation is None or snap.summary is None \
                    or str_as_bool(snap.summary.get(SnapshotSummary.PARTITION_SUMMARY_PROP, "false")):
                return None

            for key, _ in snap.summary.items():
                if key.startswith(SnapshotSummary.CHANGED_PARTITION_PREFIX):
                    part_key = key[len(SnapshotSummary.CHANGED_PARTITION_PREFIX):]
                    # part = dict(entry.split("=") for entry in val.split(","))
                    # UPDATE THIS BEFORE FINISHING
                    added_files = 0
                    added_records = 0
                    added_size = 0
                    top_n.update(part_key,
                                 lambda metrics: ((PartitionMetrics() if metrics is None else metrics)
                                                  .update_from_counts(added_files,
                                                                      added_records,
                                                                      added_size,
                                                                      snap.timestamp_millis)))

        return top_n.get()

    @staticmethod
    def snapshots_in_time_range(meta, min_ts, max_ts):
        snapshots = []
        current = meta.current_snapshot()
        while current is not None and current.timestamp_millis >= min_ts:
            current = meta.snapshot(current.parent_id)

        if current.timestamp_millis <= max_ts:
            snapshots.add(current)

        snapshots.reverse()
        return snapshots

    @staticmethod
    def timestamp_range(time_filters):
        min_timestamp = float('-inf')
        max_timestamp = float('inf')

        for pred in time_filters:
            value = pred.lit.val
            try:
                min_timestamp, max_timestamp = TIMESTAMP_RANGE_MAP[pred.op](min_timestamp, max_timestamp, value)
            except KeyError:
                raise RuntimeError("Cannot filter timestamps using predicate: %s" % pred)

        if max_timestamp < min_timestamp:
            raise RuntimeError("No timestamps can match filters: %s" % ", ".join([str(pred)
                                                                                  for pred in time_filters]))

        return min_timestamp, max_timestamp

    @staticmethod
    def join_filters(expressions):
        result = Expressions.always_true()
        for expression in expressions:
            result = Expressions.and_(result, expression)

        return result

    @staticmethod
    def to_millis(timestamp):
        if timestamp < 10000000000:
            # in seconds
            return timestamp * 1000
        elif timestamp < 10000000000000:
            # in millis
            return timestamp

        # in micros
        return timestamp / 1000


class TopN(object):

    def __init__(self, N, throw_if_limited, key_comparator):
        self.max_size = N
        self.throw_if_limited = throw_if_limited
        self.map = dict()
        self.key_comparator = key_comparator
        self.cut = None

    def update(self, key, update_func):
        if self.cut is not None and self.key_comparator(self.cut, key) <= 0:
            return

        self.map[key] = update_func(self.map.get(key))

        while len(map.keys()) > self.max_size:
            if self.throw_if_limited:
                raise RuntimeError("Too many matching keys: more than %s" % self.max_size)

            self.cut = sorted(self.map, key=functools.cmp_to_key(self.key_comparator))[-1]
            del self.map[self.cut]

    def get(self):
        return deepcopy(self.map)


class PartitionMetrics(object):

    def __init__(self):
        self.file_count = 0
        self.record_count = 0
        self.total_size = 0
        self.data_timestamp_millis = None

    def update_from_counts(self, file_count, record_count, files_size, timestamp_millis):
        self.file_count += file_count
        self.record_count += record_count
        self.total_size += files_size

        if self.data_timestamp_millis is None or self.data_timestamp_millis < timestamp_millis:
            self.data_timestamp_millis = timestamp_millis

        return self

    def update_from_file(self, file, timestamp_millis):
        self.file_count += 1
        self.record_count += file.record_count()
        self.total_size += file.files_size_in_bytes()

        if self.data_timestamp_millis is None or self.data_timestamp_millis < timestamp_millis:
            self.data_timestamp_millis = timestamp_millis

        return self

    def __repr__(self):
        items = ("%s=%r" % (k, v) for k, v in self.__dict__.items())
        return "%s(%s)" % (self.__class__.__name__, ','.join(items))

    def __str__(self):
        return self.__repr__()


class SnapshotSummary(object):
    GENIE_ID_PROP = "genie-id"
    ADDED_FILES_PROP = "added-data-files"
    DELETED_FILES_PROP = "deleted-data-files"
    TOTAL_FILES_PROP = "total-data-files"
    ADDED_RECORDS_PROP = "added-records"
    DELETED_RECORDS_PROP = "deleted-records"
    TOTAL_RECORDS_PROP = "total-records"
    ADDED_FILE_SIZE_PROP = "added-files-size"
    DELETED_DUPLICATE_FILES = "deleted-duplicate-files"
    CHANGED_PARTITION_COUNT_PROP = "changed-partition-count"
    CHANGED_PARTITION_PREFIX = "partitions."
    PARTITION_SUMMARY_PROP = "partition-summaries-included"

    def __init__(self):
        pass


class SnapshotSummaryBuilder(object):

    def __init__(self):
        self.changed_partitions = dict()
        self.added_files = 0
        self.deleted_files = 0
        self.deleted_dupicate_files = 0
        self.added_records = 0
        self.deleted_records = 0
        self.properties = dict()

    def clear(self):
        self.changed_partitions = dict()
        self.added_files = 0
        self.deleted_files = 0
        self.deleted_dupicate_files = 0
        self.added_records = 0
        self.deleted_records = 0

    def increment_duplicate_deletes(self):
        self.deleted_dupicate_files += 1

    def deleted_file(self, spec, data_file):
        self.update_partitiomns(spec, data_file, False)
        self.deleted_files += 1
        self.deleted_records += data_file.record_count

    def added_file(self, spec, data_file):
        self.update_partitiomns(spec, data_file, True)
        self.added_files += 1
        self.added_records += data_file.record_count

    def update_partitions(self, spec, file, is_addition):
        key = spec.partition_to_path(file.partition())
        metrics = self.changed_partitions.get(key, PartitionMetrics())

        if is_addition:
            self.changed_partitions[key] = metrics.update_from_file(file, None)

    def build(self):
        builder = dict()
        builder.update(self.properties)

        SnapshotSummaryBuilder.set_if(self.added_files > 0, builder,
                                      SnapshotSummary.ADDED_FILES_PROP, self.added_files)
        SnapshotSummaryBuilder.set_if(self.deleted_files > 0,
                                      builder, SnapshotSummary.DELETED_FILES_PROP, self.deleted_files)
        SnapshotSummaryBuilder.set_if(self.deleted_dupicate_files > 0,
                                      builder, SnapshotSummary.DELETED_DUPLICATE_FILES, self.deleted_dupicate_files)
        SnapshotSummaryBuilder.set_if(self.added_records > 0,
                                      builder, SnapshotSummary.ADDED_RECORDS_PROP, self.added_records)
        SnapshotSummaryBuilder.set_if(self.deleted_records > 0,
                                      builder, SnapshotSummary.DELETED_RECORDS_PROP, self.deleted_records)
        builder[SnapshotSummary.CHANGED_PARTITION_COUNT_PROP] = len(self.changed_partitions.items())

        if len(self.changed_partitions.items()) < 100:
            builder[SnapshotSummary.PARTITION_SUMMARY_PROP] = "true"
            for key, metrics in self.changed_partitions:
                metric_dict = {SnapshotSummary.ADDED_FILES_PROP: metrics.file_count,
                               SnapshotSummary.ADDED_RECORDS_PROP: metrics.record_count,
                               SnapshotSummary.ADDED_FILE_SIZE_PROP: metrics.total_size}
                builder[SnapshotSummary.CHANGED_PARTITION_PREFIX + key] = ",".join(["{}={}".format(key, val)
                                                                                    for inner_key, val
                                                                                    in metric_dict.items()])

        return builder

    def set(self, prop, value):
        self.properties[prop] = value

    @staticmethod
    def set_if(expression, builder, prop, value):
        if expression:
            builder[prop] = value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from datetime import datetime, timedelta

import pytz


class TransformUtil(object):
    EPOCH = datetime.utcfromtimestamp(0)
    EPOCH_YEAR = datetime.utcfromtimestamp(0).year

    @staticmethod
    def human_year(year_ordinal):
        return "{0:0=4d}".format(TransformUtil.EPOCH_YEAR + year_ordinal)

    @staticmethod
    def human_month(month_ordinal):
        return "{0:0=4d}-{1:0=2d}".format(TransformUtil.EPOCH_YEAR + int(month_ordinal / 12), 1 + int(month_ordinal % 12))

    @staticmethod
    def human_day(day_ordinal):
        day = TransformUtil.EPOCH + timedelta(days=day_ordinal)
        return "{0:0=4d}-{1:0=2d}-{2:0=2d}".format(day.year, day.month, day.day)

    @staticmethod
    def human_time(micros_from_midnight):
        day = TransformUtil.EPOCH + timedelta(microseconds=micros_from_midnight)
        return "{}".format(day.time())

    @staticmethod
    def human_timestamp_with_timezone(timestamp_micros):
        day = TransformUtil.EPOCH + timedelta(microseconds=timestamp_micros)
        return pytz.timezone("UTC").localize(day).strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    @staticmethod
    def human_timestamp_without_timezone(timestamp_micros):
        day = TransformUtil.EPOCH + timedelta(microseconds=timestamp_micros)
        return day.isoformat()

    @staticmethod
    def human_hour(hour_ordinal):
        time = TransformUtil.EPOCH + timedelta(hours=hour_ordinal)
        return "{0:0=4d}-{1:0=2d}-{2:0=2d}-{3:0=2d}".format(time.year, time.month, time.day, time.hour)

    @staticmethod
    def base_64_encode(buffer):
        raise NotImplementedError()

    @staticmethod
    def diff_hour(date1, date2):
        return int((date1 - date2).total_seconds() / 3600)

    @staticmethod
    def diff_day(date1, date2):
        return (date1 - date2).days

    @staticmethod
    def diff_month(date1, date2):
        return (date1.year - date2.year) * 12 + (date1.month - date2.month) - (1 if date1.day < date2.day else 0)

    @staticmethod
    def diff_year(date1, date2):
        return (date1.year - date2.year) - \
               (1 if date1.month < date2.month or (date1.month == date2.month and date1.day < date2.day) else 0)

    @staticmethod
    def unscale_decimal(decimal_value):
        value_tuple = decimal_value.as_tuple()
        return int(("-" if value_tuple.sign else "") + "".join([str(d) for d in value_tuple.digits]))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from threading import RLock


class AtomicInteger(object):

    def __init__(self, value):
        if not isinstance(value, int):
            raise RuntimeError("Value %s is not an int" % value)
        self.value = value
        self._lock = RLock()

    def increment_and_get(self):
        with self._lock:
            self.value += 1
            return self.value

    def get(self):
        with self._lock:
            return self.value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json

from .base_snapshot import BaseSnapshot


class SnapshotParser(object):
    SNAPSHOT_ID = "snapshot-id"
    PARENT_SNAPSHOT_ID = "parent-snapshot-id"
    TIMESTAMP_MS = "timestamp-ms"
    SUMMARY = "summary"
    OPERATION = "operation"
    MANIFESTS = "manifests"
    MANIFEST_LIST = "manifest-list"

    @staticmethod
    def to_json(snapshot):
        return json.dumps(SnapshotParser.to_dict(snapshot))

    @staticmethod
    def to_dict(snapshot):
        return {SnapshotParser.SNAPSHOT_ID: snapshot.snapshot_id,
                SnapshotParser.TIMESTAMP_MS: snapshot.timestamp_millis,
                SnapshotParser.PARENT_SNAPSHOT_ID: snapshot.parent_id,
                SnapshotParser.OPERATION: snapshot.operation,
                SnapshotParser.SUMMARY: snapshot.summary,
                SnapshotParser.MANIFESTS: [manifest.manifest_path for manifest in snapshot.manifests]}

    @staticmethod
    def from_json(ops, json_obj, spec=None):
        if isinstance(json_obj, str):
            json_obj = json.loads(json_obj)

        if not isinstance(json_obj, dict):
            raise RuntimeError("Cannot parse table version from a non-object: %s" % json_obj)

        version_id = json_obj.get(SnapshotParser.SNAPSHOT_ID)
        parent_id = json_obj.get(SnapshotParser.PARENT_SNAPSHOT_ID)
        timestamp_millis = json_obj.get(SnapshotParser.TIMESTAMP_MS)
        operation = json_obj.get(SnapshotParser.OPERATION)
        summary = json_obj.get(SnapshotParser.SUMMARY)

        if SnapshotParser.MANIFEST_LIST in json_obj:
            # the manifest list is stored in a manifest list file
            return BaseSnapshot(ops, version_id, parent_id,
                                manifest_list=ops.new_input_file(json_obj.get(SnapshotParser.MANIFEST_LIST)),
                                timestamp_millis=timestamp_millis,
                                operation=operation,
                                summary=summary)
        else:
            manifests = json_obj.get(SnapshotParser.MANIFESTS, list())

            return BaseSnapshot(ops, version_id, parent_id,
                                manifests=manifests,
                                timestamp_millis=timestamp_millis,
                                operation=operation,
                                summary=summary)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from iceberg.api import PartitionFieldSummary, StructLike


class GenericPartitionFieldSummary(PartitionFieldSummary, StructLike):

    AVRO_SCHEMA = None  # IcebergToAvro.type_to_schema(PartitionFieldSummary.get_type())

    def __init__(self, avro_schema=None, contains_null=False, lower_bound=None, upper_bound=None, copy=None):
        if copy is not None:
            avro_schema = copy.avro_schema
            contains_null = copy.contains_null()
            lower_bound = copy.lower_bound()
            upper_bound = copy.upper_bound()

        if avro_schema is None:
            avro_schema = GenericPartitionFieldSummary.AVRO_SCHEMA

        self.avro_schema = avro_schema
        self._contains_null = contains_null
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def __str__(self):
        return ("GenericPartitionFieldSummary(contains_null={},lower_bound={}, upper_bound={})"
                .format(self.contains_null(), self.lower_bound(), self.upper_bound()))

    def contains_null(self):
        return self._contains_null

    def get(self, pos):
        raise NotImplementedError()

    def set(self, pos, value):
        raise NotImplementedError()

    def lower_bound(self):
        return self._lower_bound

    def upper_bound(self):
        return self._upper_bound

    def copy(self):
        return GenericPartitionFieldSummary(copy=self)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

__all__ = ["ABOVE_MAX",
           "And",
           "BELOW_MIN",
           "BinaryLiteral",
           "Binder",
           "BooleanLiteral",
           "BoundPredicate",
           "BoundReference",
           "Evaluator",
           "Expression",
           "ExpressionVisitors",
           "Expressions",
           "DateLiteral",
           "DecimalLiteral",
           "DoubleLiteral",
           "FALSE",
           "FalseExp",
           "FixedLiteral",
           "FloatLiteral",
           "inclusive",
           "InclusiveManifestEvaluator",
           "InclusiveMetricsEvaluator",
           "InclusiveProjection",
           "IntegerLiteral",
           "JAVA_MAX_FLOAT",
           "JAVA_MAX_INT",
           "JAVA_MIN_FLOAT",
           "JAVA_MIN_INT",
           "Literal",
           "Literals",
           "NamedReference",
           "Not",
           "Operation",
           "Or",
           "Predicate",
           "ResidualEvaluator",
           "strict",
           "StrictMetricsEvaluator",
           "StrictProjection",
           "StringLiteral",
           "TRUE",
           "TrueExp",
           "UUIDLiteral",
           "UnboundPredicate"]

from .evaluator import (Binder,
                        Evaluator)
from .expression import (And,
                         Expression,
                         FALSE,
                         FalseExp,
                         Not,
                         Operation,
                         Or,
                         TRUE,
                         TrueExp)
from .expressions import Expressions, ExpressionVisitors
from .inclusive_manifest_evaluator import InclusiveManifestEvaluator
from .inclusive_metrics_evaluator import InclusiveMetricsEvaluator
from .java_variables import (JAVA_MAX_FLOAT,
                             JAVA_MAX_INT,
                             JAVA_MIN_FLOAT,
                             JAVA_MIN_INT)
from .literals import (ABOVE_MAX,
                       BELOW_MIN,
                       BinaryLiteral,
                       BooleanLiteral,
                       DateLiteral,
                       DecimalLiteral,
                       DoubleLiteral,
                       FixedLiteral,
                       FloatLiteral,
                       IntegerLiteral,
                       Literal,
                       Literals,
                       StringLiteral,
                       UUIDLiteral)
from .predicate import (BoundPredicate,
                        Predicate,
                        UnboundPredicate)
from .projections import (inclusive,
                          InclusiveProjection,
                          strict,
                          StrictProjection)
from .reference import (BoundReference,
                        NamedReference)
from .residual_evaluator import ResidualEvaluator
from .strict_metrics_evaluator import StrictMetricsEvaluator



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from iceberg.api.types.conversions import Conversions

from .generic_partition_field_summary import GenericPartitionFieldSummary


class PartitionSummary(object):

    def __init__(self, spec):
        self.java_classes = spec.java_classes
        self.fields = [PartitionFieldStats(spec.partition_type.fields[i].type)
                       for i in range(len(self.java_classes))]

    def summaries(self):
        return [field.to_summary() for field in self.fields]

    def update(self, partition_key):
        self.update_fields(partition_key)

    def update_fields(self, key):
        for i in range(self.java_classes):
            stats = self.fields[i]
            java_class = self.java_classes[i]
            stats.update(key.get(i, java_class))


class PartitionFieldStats(object):

    def __init__(self, type_var):
        self.contains_null = False
        self.type = type_var
        self.min = None
        self.max = None

    def to_summary(self):
        lower_bound = None if self.min is None else Conversions.to_byte_buffer(self.type.type_id, self.min)
        upper_bound = None if self.max is None else Conversions.to_byte_buffer(self.type.type_id, self.max)
        return GenericPartitionFieldSummary(contains_null=self.contains_null,
                                            lower_bound=lower_bound,
                                            upper_bound=upper_bound)

    def update(self, value):
        if value is None:
            self.contains_null = True
        elif self.min is None:
            self.min = value
            self.max = value
        else:
            if value < self.min:
                self.min = value
            if value > self.max:
                self.max = value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import threading

from .binder import Binder
from .expressions import ExpressionVisitors


class Evaluator(object):

    def __init__(self, struct, unbound, case_sensitive=True):
        self.expr = Binder.bind(struct, unbound, case_sensitive)
        self.thread_local_data = threading.local()

    def _visitor(self):
        if not hasattr(self.thread_local_data, "visitors"):
            self.thread_local_data.visitors = Evaluator.EvalVisitor()

        return self.thread_local_data.visitors

    def eval(self, data):
        return self._visitor().eval(data, self.expr)

    class EvalVisitor(ExpressionVisitors.BoundExpressionVisitor):

        def __init__(self):
            super(Evaluator.EvalVisitor, self).__init__()
            self.struct = None

        def eval(self, row, expr):
            self.struct = row
            return ExpressionVisitors.visit(expr, self)

        def always_true(self):
            return True

        def always_false(self):
            return False

        def not_(self, result):
            return not result

        def and_(self, left_result, right_result):
            return left_result and right_result

        def or_(self, left_result, right_result):
            return left_result or right_result

        def is_null(self, ref):
            return ref.get(self.struct) is None

        def not_null(self, ref):
            return not (ref.get(self.struct) is None)

        def lt(self, ref, lit):
            return ref.get(self.struct) < lit.value

        def lt_eq(self, ref, lit):
            return ref.get(self.struct) <= lit.value

        def gt(self, ref, lit):
            return ref.get(self.struct) > lit.value

        def gt_eq(self, ref, lit):
            return ref.get(self.struct) >= lit.value

        def eq(self, ref, lit):
            return ref.get(self.struct) == lit.value

        def not_eq(self, ref, lit):
            return ref.get(self.struct) != lit.value

        def in_(self, ref, lit):
            raise NotImplementedError()

        def not_in(self, ref, lit):
            return not self.in_(ref, lit.value)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from iceberg.api.types import StructType

if TYPE_CHECKING:
    from iceberg.api import StructLike


class Bound(object):
    def eval(self, struct: StructLike) -> Any:
        """Evaluate current object against struct"""


class Unbound(object):
    @property
    def ref(self):
        """the underlying reference"""

    def bind(self, struct: StructType, case_sensitive: bool = True) -> Bound:
        """bind the current object based on the given struct and return the Bound object"""


class Term(object):
    @property
    def ref(self):
        """Expose the reference for this term"""

    @property
    def type(self):
        """accessor for term type """


class BoundTerm(Bound, Term):

    @property
    def ref(self):
        raise NotImplementedError("Base class does not have implementation")

    @property
    def type(self):
        raise NotImplementedError("Base class does not have implementation")

    def eval(self, struct: StructLike):
        raise NotImplementedError("Base class does not have implementation")


class UnboundTerm(Unbound, Term):

    @property
    def ref(self):
        raise NotImplementedError("Base class does not have implementation")

    def bind(self, struct: StructType, case_sensitive: bool = True):
        raise NotImplementedError("Base class does not have implementation")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import gzip

from iceberg.api.io import InputFile, OutputFile

from .util import get_fs


class FileSystem(object):

    def open(self, path, mode='rb'):
        raise NotImplementedError()

    def create(self, path, overwrite=False):
        raise NotImplementedError()

    def exists(self, path):
        raise NotImplementedError()

    def delete(self, path):
        raise NotImplementedError()

    def stat(self, path):
        raise NotImplementedError()

    def rename(self, src, dest):
        raise NotImplementedError()


class FileSystemInputFile(InputFile):

    def __init__(self, fs, path, conf, length=None, stat=None):
        self.fs = fs
        self.path = path
        self.conf = conf
        self.length = length
        self.stat = stat

    @staticmethod
    def from_location(location, conf):
        fs = get_fs(location, conf)
        return FileSystemInputFile(fs, location, conf)

    def location(self):
        return self.path

    def get_length(self):
        return self.get_stat().length

    def get_stat(self):
        return self.lazy_stat()

    def lazy_stat(self):
        if self.stat is None:
            self.stat = self.fs.stat(self.path)
        return self.stat

    def new_stream(self, gzipped=False):
        with self.fs.open(self.location()) as fo:
            if gzipped:
                fo = gzip.GzipFile(fileobj=fo)
            for line in fo:
                yield line

    def new_fo(self, mode="rb"):
        return self.fs.open(self.location(), mode=mode)

    def __repr__(self):
        return "FileSystemInputFile({})".format(self.path)

    def __str__(self):
        return self.__repr__()


class FileSystemOutputFile(OutputFile):

    @staticmethod
    def from_path(path, conf):
        return FileSystemOutputFile(path, conf)

    def __init__(self, path, conf):
        self.path = path
        self.conf = conf

    def create(self, mode="w"):
        fs = get_fs(self.path, self.conf)
        if fs.exists(self.path):
            raise RuntimeError("File %s already exists" % self.path)

        return fs.open(self.path, mode=mode)

    def create_or_overwrite(self):
        fs = get_fs(self.path, self.conf)

        return fs.open(self.path, "wb")

    def location(self):
        return str(self.path)

    def __repr__(self):
        return "FileSystemOutputFile({})".format(self.path)

    def __str__(self):
        return self.__repr__()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from iceberg.exceptions import ValidationException

from ..types import StructType

if TYPE_CHECKING:
    from iceberg.api import StructLike


class BoundReference:

    def __init__(self, struct, field):
        self.field = field
        self.pos = self.find(field.field_id, struct)
        self._type = struct.fields[self.pos].type

    @property
    def type(self):
        return self._type

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, BoundReference):
            return False

        return self.field.field_id == other.field.field_id and self.pos == other.pos and self._type == other._type

    def __ne__(self, other):
        return not self.__eq__(other)

    def find(self, field_id, struct):
        fields = struct.fields
        for i, field in enumerate(fields):
            if field.field_id == self.field.field_id:
                return i

        raise ValidationException("Cannot find top-level field id %d in struct: %s", (field_id, struct))

    def get(self, struct):
        return struct.get(self.pos)

    def __str__(self):
        return "ref(id={id}, pos={pos}, type={_type})".format(id=self.field.field_id,
                                                              pos=self.pos,
                                                              _type=self._type)

    @property
    def ref(self):
        return self

    def eval(self, struct: StructLike) -> Any:
        return self.get(struct)


class NamedReference:

    def __init__(self, name):
        super(NamedReference, self).__init__()
        if name is None:
            raise RuntimeError("Name cannot be null")

        self.name = name

    @property
    def ref(self):
        return self

    def bind(self, struct: StructType, case_sensitive: bool = True) -> BoundReference:
        from iceberg.api import Schema
        schema = Schema(struct.fields)
        field = schema.find_field(self.name) if case_sensitive else schema.case_insensitive_find_field(self.name)

        ValidationException.check(field is not None, "Cannot find field '%s' in struct: %s", (self.name,
                                                                                              schema.as_struct()))

        return BoundReference(struct, field)

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, NamedReference):
            return False

        return self.name == other.name

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "NamedReference({})".format(self.name)

    def __str__(self):
        return 'ref(name="{}")'.format(self.name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from pathlib import Path
from typing import Iterator
import uuid

from iceberg.exceptions import CommitFailedException, ValidationException

from .file_system import FileSystemInputFile, FileSystemOutputFile
from .util import get_fs
from ..table_metadata_parser import TableMetadataParser
from ..table_operations import TableOperations
from ..table_properties import TableProperties

_logger = logging.getLogger(__name__)


class FilesystemTableOperations(TableOperations):

    def __init__(self, location, conf):
        self.conf = conf
        self.location = Path(location)
        self.should_refresh = True
        self.version = None
        self.current_metadata = None

    def current(self):
        if self.should_refresh:
            return self.refresh()

        return self.current_metadata

    def refresh(self):
        ver = self.version if self.version is not None else self.read_version_hint()
        metadata_file = self.metadata_file(ver)
        fs = get_fs(str(metadata_file), self.conf)

        if ver is not None and not fs.exists(metadata_file):
            if ver == 0:
                return None
            raise ValidationException("Metadata file is missing: %s" % metadata_file)

        while fs.exists(str(self.metadata_file(ver + 1))):
            ver += 1
            metadata_file = self.metadata_file(ver)

        self.version = ver
        self.current_metadata = TableMetadataParser.read(self, FileSystemInputFile.from_location(str(metadata_file),
                                                                                                 self.conf))
        self.should_refresh = False
        return self.current_metadata

    def commit(self, base, metadata):
        if base != self.current():
            raise CommitFailedException("Cannot commit changes based on stale table metadata")

        if not (base is None or base.location() == metadata.location()):
            raise RuntimeError("Hadoop path-based tables cannot be relocated")
        if TableProperties.WRITE_METADATA_LOCATION in metadata.properties:
            raise RuntimeError("Hadoop path-based tables cannot be relocated")

        temp_metadata_file = self.metadata_path("{}{}".format(uuid.uuid4(),
                                                              TableMetadataParser.get_file_extension(self.conf)))
        TableMetadataParser.write(metadata, FileSystemOutputFile.from_path(str(temp_metadata_file), self.conf))

        next_version = (self.version if self.version is not None else 0) + 1
        final_metadata_file = self.metadata_file(next_version)
        fs = get_fs(str(final_metadata_file), self.conf)

        if fs.exists(final_metadata_file):
            raise CommitFailedException("Version %s already exists: %s" % (next_version, final_metadata_file))

        if not fs.rename(temp_metadata_file, final_metadata_file):
            raise CommitFailedException("Failed to commit changes using rename: %s" % final_metadata_file)

        self.write_version_hint(next_version)
        self.should_refresh = True

    def new_input_file(self, path):
        return FileSystemInputFile.from_location(path, self.conf)

    def new_output_file(self, path):
        return FileSystemOutputFile.from_path(path, self.conf)

    def new_metadata_file(self, filename):
        raise RuntimeError("Not yet implemented")

    def delete_file(self, path):
        get_fs(path, self.conf).delete(path)

    def new_snapshot_id(self):
        raise RuntimeError("Not yet implemented")

    def metadata_file_location(self, file):
        return str(self.metadata_path(file))

    def metadata_file(self, version):
        return self.metadata_path("v{}{}".format(version, TableMetadataParser.get_file_extension(self.conf)))

    def metadata_path(self, filename):
        return self.location / Path("metadata") / Path(filename)

    def version_hint_file(self):
        return self.metadata_path("version-hint.text")

    def read_version_hint(self):
        version_hint_file = str(self.version_hint_file())
        fs = get_fs(version_hint_file, self.conf)

        if not fs.exists(version_hint_file):
            return 0
        else:
            with fs.open(version_hint_file, "r") as fo:
                # return int(fo.readline().replace("\n", ""))
                line = fo.readline()
                if isinstance(line, Iterator):
                    return int(next(line))
                return int(line)

    def write_version_hint(self, version):
        version_hint_file = str(self.version_hint_file())
        fs = get_fs(version_hint_file, self.conf)
        try:

            with fs.create(version_hint_file, True) as fo:
                fo.write("{}".format(version))

        except RuntimeError as e:
            _logger.warning("Unable to update version hint: %s" % e)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from decimal import Decimal

from .projection_util import ProjectionUtil
from .transform import Transform
from .transform_util import TransformUtil
from ..expressions import (Expressions,
                           Operation)
from ..types import TypeID


class Truncate(Transform):

    @staticmethod
    def get(type_var, width):
        if type_var.type_id == TypeID.INTEGER:
            return TruncateInteger(width)
        elif type_var.type_id == TypeID.LONG:
            return TruncateLong(width)
        elif type_var.type_id == TypeID.DECIMAL:
            return TruncateDecimal(width)
        elif type_var.type_id == TypeID.STRING:
            return TruncateString(width)

    def __init__(self):
        raise NotImplementedError()

    def apply(self, value):
        raise NotImplementedError()

    def can_transform(self, type_var):
        raise NotImplementedError()

    def get_result_type(self, source_type):
        return source_type

    def project(self, name, predicate):
        raise NotImplementedError()

    def project_strict(self, name, predicate):
        raise NotImplementedError()


class TruncateInteger(Truncate):

    def __init__(self, width):
        self.W = width

    def apply(self, value):
        return value - (((value % self.W) + self.W) % self.W)

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.INTEGER

    def project(self, name, predicate):
        if predicate.op == Operation.NOT_NULL or predicate.op == Operation.IS_NULL:
            return Expressions.predicate(predicate.op, name)

        return ProjectionUtil.truncate_integer(name, predicate, self)

    def project_strict(self, name, predicate):
        if predicate.op == Operation.LT:
            _in = predicate.lit.value - 1
            _out = predicate.lit.value
            in_image = self.apply(_in)
            out_image = self.apply(_out)
            if in_image != out_image:
                return Expressions.predicate(Operation.LT_EQ, name, in_image)
            else:
                return Expressions.predicate(Operation.LT, name, in_image)
        elif predicate.op == Operation.LT_EQ:
            _in = predicate.lit.value
            _out = predicate.lit.value + 1
            in_image = self.apply(_in)
            out_image = self.apply(_out)
            if in_image != out_image:
                return Expressions.predicate(Operation.LT_EQ, name, in_image)
            else:
                return Expressions.predicate(Operation.LT, name, in_image)

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, TruncateInteger):
            return False

        return self.W == other.W

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return TruncateInteger.__class__, self.W

    def __str__(self):
        return "truncate[%s]" % self.W


class TruncateLong(Truncate):

    def __init__(self, width):
        self.W = width

    def apply(self, value):
        return value - (((value % self.W) + self.W) % self.W)

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.LONG

    def project(self, name, predicate):
        if predicate.op == Operation.NOT_NULL or predicate.op == Operation.IS_NULL:
            return Expressions.predicate(predicate.op, name)

        return ProjectionUtil.truncate_long(name, predicate, self)

    def project_strict(self, name, predicate):
        return None

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, TruncateLong):
            return False

        return self.W == other.W

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return TruncateLong.__class__, self.W

    def __str__(self):
        return "truncate[%s]" % self.W


class TruncateDecimal(Truncate):

    def __init__(self, unscaled_width):
        self.unscaled_width = unscaled_width

    def apply(self, value):
        unscaled_value = TransformUtil.unscale_decimal(value)
        applied_value = unscaled_value - (((unscaled_value % self.unscaled_width) + self.unscaled_width) % self.unscaled_width)
        return Decimal("{}e{}".format(applied_value, value.as_tuple().exponent))

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.DECIMAL

    def project(self, name, predicate):
        if predicate.op == Operation.NOT_NULL or predicate.op == Operation.IS_NULL:
            return Expressions.predicate(predicate.op, name)

        return ProjectionUtil.truncate_decimal(name, predicate, self)

    def project_strict(self, name, predicate):
        return None

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, TruncateDecimal):
            return False

        return self.unscaled_width == other.unscaled_width

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return TruncateDecimal.__class__, self.unscaled_width

    def __str__(self):
        return "truncate[%s]" % self.unscaled_width


class TruncateString(Truncate):
    def __init__(self, length):
        self.L = length

    def apply(self, value):
        return value[0:min(self.L, len(value))]

    def can_transform(self, type_var):
        return type_var.type_id == TypeID.STRING

    def project(self, name, predicate):
        if predicate.op == Operation.NOT_NULL or predicate.op == Operation.IS_NULL:
            return Expressions.predicate(predicate.op, name)

        return ProjectionUtil.truncate_array(name, predicate, self)

    def project_strict(self, name, predicate):
        return None

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, TruncateString):
            return False

        return self.L == other.L

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return TruncateString.__class__, self.L

    def __str__(self):
        return "truncate[%s]" % self.L



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .expressions import Expressions
from .filterable import Filterable


class FilteredSnapshot(Filterable):

    def __init__(self, snapshot, part_filter, row_filter, columns):
        self.snapshot = snapshot
        self.part_filter = part_filter
        self.row_filter = row_filter
        self.columns = columns

    def select(self, columns):
        return FilteredSnapshot(self.snapshot, self.part_filter, self.row_filter, columns)

    def filter_partitions(self, expr):
        return FilteredSnapshot(self.snapshot, Expressions.and_(self.part_filter, expr), self.row_filter, self.columns)

    def filter_rows(self, expr):
        return FilteredSnapshot(self.snapshot, self.part_filter, Expressions.and_(self.row_filter, expr), self.columns)

    def iterator(self):
        return self.snapshot.iterator(self.part_filter, self.row_filter, self.columns)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import fastavro
from iceberg.api import Schema
from iceberg.api.types import (BinaryType,
                               BooleanType,
                               DateType,
                               DoubleType,
                               FloatType,
                               IntegerType,
                               ListType,
                               LongType,
                               MapType,
                               NestedField,
                               StringType,
                               StructType,
                               TimestampType,
                               TimeType,
                               TypeID)


class AvroToIceberg(object):
    FIELD_ID_PROP = "field-id"
    FIELD_TYPE_PROP = "type"
    FIELD_NAME_PROP = "name"
    FIELD_LOGICAL_TYPE_PROP = "logicalType"
    FIELD_FIELDS_PROP = "fields"
    FIELD_ITEMS_PROP = "items"
    FIELD_ELEMENT_ID_PROP = "element-id"

    AVRO_JSON_PRIMITIVE_TYPES = ("boolean", "int", "long", "float", "double", "bytes", "string")
    AVRO_JSON_COMPLEX_TYPES = ("record", "array", "enum", "fixed")

    TYPE_PROCESSING_MAP = {str: lambda x, y: AvroToIceberg.convert_str_type(x, y),
                           dict: lambda x, y: AvroToIceberg.convert_complex_type(x, y),
                           list: lambda x, y: AvroToIceberg.convert_union_type(x, y)}

    COMPLEX_TYPE_PROCESSING_MAP = {"record": lambda x, y: AvroToIceberg.convert_record_type(x, y),
                                   "array": lambda x, y: AvroToIceberg.convert_array_type(x, y),
                                   "map": lambda x, y: AvroToIceberg.convert_map_type(x, y)}

    PRIMITIVE_FIELD_TYPE_MAP = {"boolean": BooleanType.get(),
                                "bytes": BinaryType.get(),
                                "date": DateType.get(),
                                "double": DoubleType.get(),
                                "float": FloatType.get(),
                                "int": IntegerType.get(),
                                "long": LongType.get(),
                                "string": StringType.get(),
                                "time-millis": TimeType.get(),
                                "timestamp-millis": TimestampType.without_timezone()}

    PROCESS_FUNCS = {TypeID.STRUCT: lambda avro_row, field: AvroToIceberg.get_field_from_struct(avro_row, field),
                     TypeID.LIST: lambda avro_row, field: AvroToIceberg.get_field_from_list(avro_row, field),
                     TypeID.MAP: lambda avro_row, field: AvroToIceberg.get_field_from_map(avro_row, field)}

    @staticmethod
    def convert_avro_schema_to_iceberg(avro_schema):
        if avro_schema.get(AvroToIceberg.FIELD_TYPE_PROP) != "record":
            raise RuntimeError("Cannot convert avro schema to iceberg %s" % avro_schema)

        struct = AvroToIceberg.convert_type(avro_schema, None)

        return Schema(struct[0].fields)

    @staticmethod
    def convert_record_type(avro_field, next_id=None):
        avro_field_type = avro_field.get(AvroToIceberg.FIELD_TYPE_PROP)

        if avro_field_type != "record":
            raise RuntimeError("Field type muse be 'record': %s" % avro_field_type)

        fields = avro_field.get(AvroToIceberg.FIELD_FIELDS_PROP)

        iceberg_fields = []
        if next_id is None:
            next_id = len(fields)
        for field in fields:
            iceberg_field, next_id = AvroToIceberg.convert_avro_field_to_iceberg(field, next_id=next_id)
            iceberg_fields.append(iceberg_field)

        return StructType.of(iceberg_fields), next_id

    @staticmethod
    def convert_avro_field_to_iceberg(field, next_id):
        field_type, is_optional, next_id = AvroToIceberg.convert_type(field, next_id)

        if field.get(AvroToIceberg.FIELD_ID_PROP) is None:
            return field_type, next_id

        if is_optional:
            return NestedField.optional(field.get(AvroToIceberg.FIELD_ID_PROP),
                                        field.get(AvroToIceberg.FIELD_NAME_PROP),
                                        field_type), next_id
        else:
            return NestedField.required(field.get(AvroToIceberg.FIELD_ID_PROP),
                                        field.get(AvroToIceberg.FIELD_NAME_PROP),
                                        field_type), next_id

    @staticmethod
    def convert_type(field, next_id=None):
        avro_field_type = field.get(AvroToIceberg.FIELD_TYPE_PROP)

        optional = AvroToIceberg.is_option_schema(avro_field_type)

        processing_func = AvroToIceberg.TYPE_PROCESSING_MAP.get(type(avro_field_type))
        if processing_func is None:
            raise RuntimeError("No function found to process %s" % avro_field_type)

        iceberg_type, next_id = processing_func(field, next_id)

        return iceberg_type, optional, next_id

    @staticmethod
    def convert_str_type(avro_field, next_id=None):
        avro_field_type = avro_field.get(AvroToIceberg.FIELD_TYPE_PROP)
        logical_type = avro_field.get(AvroToIceberg.FIELD_LOGICAL_TYPE_PROP)
        if not isinstance(avro_field_type, str):
            raise RuntimeError("Field type must be of type str: %s" % avro_field_type)

        if avro_field_type in AvroToIceberg.AVRO_JSON_PRIMITIVE_TYPES:
            if logical_type is not None:
                return AvroToIceberg.PRIMITIVE_FIELD_TYPE_MAP.get(logical_type), next_id
            else:
                return AvroToIceberg.PRIMITIVE_FIELD_TYPE_MAP.get(avro_field_type), next_id

        elif avro_field_type in AvroToIceberg.AVRO_JSON_COMPLEX_TYPES:
            if logical_type is not None:
                processing_func = AvroToIceberg.COMPLEX_TYPE_PROCESSING_MAP.get(logical_type)
            else:
                processing_func = AvroToIceberg.COMPLEX_TYPE_PROCESSING_MAP.get(avro_field_type)

            if processing_func is None:
                raise RuntimeError("No function found to process %s" % avro_field_type)

            return processing_func(avro_field, next_id)
        else:
            raise RuntimeError("Unknown type %s" % avro_field_type)

    @staticmethod
    def convert_complex_type(avro_field, next_id=None):
        avro_field_type = avro_field.get(AvroToIceberg.FIELD_TYPE_PROP)
        if not isinstance(avro_field_type, dict):
            raise RuntimeError("Complex field type must be of type dict: %s" % avro_field_type)

        return AvroToIceberg.convert_avro_field_to_iceberg(avro_field_type, next_id)

    @staticmethod
    def convert_union_type(avro_field, next_id=None):
        avro_field_type = avro_field.get(AvroToIceberg.FIELD_TYPE_PROP)
        if not isinstance(avro_field_type, list):
            raise RuntimeError("Union field type must be of type list: %s" % avro_field_type)

        if len(avro_field_type) > 2:
            raise RuntimeError("Cannot process unions larger than 2 items: %s" % avro_field_type)
        for item in avro_field_type:
            if isinstance(item, str) and item == "null":
                continue
            avro_field_type = item
        avro_field[AvroToIceberg.FIELD_TYPE_PROP] = avro_field_type
        items = AvroToIceberg.convert_type(avro_field, next_id)
        return items[0], items[2]

    @staticmethod
    def convert_array_type(avro_field, next_id=None):
        avro_field_type = avro_field.get(AvroToIceberg.FIELD_TYPE_PROP)
        if avro_field_type != "array":
            raise RuntimeError("Avro type must be array: %s" % avro_field_type)
        element_id = avro_field.get(AvroToIceberg.FIELD_ELEMENT_ID_PROP)
        items = avro_field.get(AvroToIceberg.FIELD_ITEMS_PROP)

        is_optional = AvroToIceberg.is_option_schema(items)

        if isinstance(items, str) and items in AvroToIceberg.PRIMITIVE_FIELD_TYPE_MAP:
            item_type = AvroToIceberg.PRIMITIVE_FIELD_TYPE_MAP.get(items)
            if item_type is None:
                raise RuntimeError("No mapping found for type %s" % items)
        else:
            raise RuntimeError("Complex list types not yet implemented")

        if is_optional:
            return ListType.of_optional(element_id, item_type), next_id
        else:
            return ListType.of_required(element_id, item_type), next_id

    @staticmethod
    def convert_map_type(avro_field, next_id=None):
        avro_field_type = avro_field.get(AvroToIceberg.FIELD_TYPE_PROP)
        avro_logical_type = avro_field.get(AvroToIceberg.FIELD_LOGICAL_TYPE_PROP)
        if avro_field_type != "array" or avro_logical_type != "map":
            raise RuntimeError("Avro type must be array and logical type must be map: %s" % avro_logical_type)
        is_optional = False
        items = avro_field.get(AvroToIceberg.FIELD_ITEMS_PROP)
        for field in items.get(AvroToIceberg.FIELD_FIELDS_PROP, list()):
            if field.get(AvroToIceberg.FIELD_NAME_PROP) == "key":
                key_id = field.get(AvroToIceberg.FIELD_ID_PROP)
                if not isinstance(field.get(AvroToIceberg.FIELD_TYPE_PROP), str):
                    raise RuntimeError("Support for complex map keys not yet implemented")
                key_type = AvroToIceberg.PRIMITIVE_FIELD_TYPE_MAP.get(field.get(AvroToIceberg.FIELD_TYPE_PROP))
            elif field.get(AvroToIceberg.FIELD_NAME_PROP) == "value":
                value_id = field.get(AvroToIceberg.FIELD_ID_PROP)
                if not isinstance(field.get(AvroToIceberg.FIELD_TYPE_PROP), str):
                    raise RuntimeError("Support for complex map values not yet imeplemented")
                value_type = AvroToIceberg.PRIMITIVE_FIELD_TYPE_MAP.get(field.get(AvroToIceberg.FIELD_TYPE_PROP))

        if is_optional:
            return MapType.of_optional(key_id, value_id, key_type, value_type), next_id
        else:
            return MapType.of_required(key_id, value_id, key_type, value_type), next_id

    @staticmethod
    def is_option_schema(field_type):
        if isinstance(field_type, list) and len(field_type) == 2 and "null" in field_type:
            return True

        return False

    @staticmethod
    def read_avro_file(iceberg_schema, data_file):
        fo = data_file.new_fo()
        avro_reader = fastavro.reader(fo)
        for avro_row in avro_reader:
            iceberg_row = dict()
            for field in iceberg_schema.as_struct().fields:
                iceberg_row[field.name] = AvroToIceberg.get_field_from_avro(avro_row, field)
            yield iceberg_row
        fo.close()

    @staticmethod
    def read_avro_row(iceberg_schema, avro_reader):
        try:
            for avro_row in avro_reader:
                iceberg_row = dict()
                for field in iceberg_schema.as_struct().fields:
                    iceberg_row[field.name] = AvroToIceberg.get_field_from_avro(avro_row, field)
                yield iceberg_row
        except StopIteration:
            return

    @staticmethod
    def get_field_from_avro(avro_row, field):
        try:
            return AvroToIceberg.PROCESS_FUNCS.get(field.type.type_id,
                                                   AvroToIceberg.get_field_from_primitive)(avro_row, field)
        except KeyError:
            raise RuntimeError("Don't know how to get field of type: %s" % field.type.type_id)

    @staticmethod
    def get_field_from_primitive(avro_row, field):
        try:
            return avro_row[field.name]
        except KeyError:
            if field.is_required:
                raise RuntimeError("Field is required but missing in source %s\n%s:" % (field, avro_row))

    @staticmethod
    def get_field_from_struct(avro_row, field):
        field_obj = {}
        for nested_field in field.type.fields:
            field_obj[nested_field.name] = AvroToIceberg.get_field_from_avro(avro_row[field.name], nested_field)
        return field_obj

    @staticmethod
    def get_field_from_list(avro_row, field):
        try:
            return avro_row[field.name]
        except KeyError:
            if field.is_required:
                raise RuntimeError("Field is required but missing in source %s\n%s:" % (field, avro_row))

    @staticmethod
    def get_field_from_map(avro_row, field):
        val_map = dict()

        try:
            avro_value = avro_row[field.name]
        except KeyError:
            if field.is_required:
                raise RuntimeError("Field is required but missing in source %s\n%s:" % (field, avro_row))
            else:
                return None

        for val in avro_value:
            val_map[val['key']] = val['value']

        return val_map



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import datetime
from decimal import (Decimal,
                     ROUND_HALF_UP)
import uuid

import pytz


from .expression import (FALSE,
                         TRUE)
from .java_variables import (JAVA_MAX_FLOAT,
                             JAVA_MIN_FLOAT)
from ..types.conversions import Conversions
from ..types.type import TypeID


class Literals(object):

    EPOCH = datetime.datetime.utcfromtimestamp(0)
    EPOCH_DAY = EPOCH.date()

    @staticmethod
    def from_(value):  # noqa: C901
        if value is None:
            raise RuntimeError("Cannot create an expression literal from None")
        if isinstance(value, bool):
            return BooleanLiteral(value)
        elif isinstance(value, int):
            if Literal.JAVA_MIN_INT < value < Literal.JAVA_MAX_INT:
                return IntegerLiteral(value)
            return LongLiteral(value)
        elif isinstance(value, float):
            if Literal.JAVA_MIN_FLOAT < value < Literal.JAVA_MAX_FLOAT:
                return FloatLiteral(value)
            return DoubleLiteral(value)
        elif isinstance(value, str):
            return StringLiteral(value)
        elif isinstance(value, uuid.UUID):
            return UUIDLiteral(value)
        elif isinstance(value, bytearray):
            return BinaryLiteral(value)
        elif isinstance(value, bytes):
            return FixedLiteral(value)
        elif isinstance(value, Decimal):
            return DecimalLiteral(value)
        else:
            raise NotImplementedError("Unimplemented Type Literal for value: %s" % value)

    @staticmethod
    def above_max():
        return ABOVE_MAX

    @staticmethod
    def below_min():
        return BELOW_MIN


class Literal(object):
    JAVA_MAX_INT = 2147483647
    JAVA_MIN_INT = -2147483648
    JAVA_MAX_FLOAT = 3.4028235E38
    JAVA_MIN_FLOAT = -3.4028235E38

    @staticmethod
    def of(value):  # noqa: C901

        if isinstance(value, bool):
            return BooleanLiteral(value)
        elif isinstance(value, int):
            if value < Literal.JAVA_MIN_INT or value > Literal.JAVA_MAX_INT:
                return LongLiteral(value)
            return IntegerLiteral(value)
        elif isinstance(value, float):
            if value < Literal.JAVA_MIN_FLOAT or value > Literal.JAVA_MAX_FLOAT:
                return DoubleLiteral(value)
            return FloatLiteral(value)
        elif isinstance(value, str):
            return StringLiteral(value)
        elif isinstance(value, uuid.UUID):
            return UUIDLiteral(value)
        elif isinstance(value, bytes):
            return FixedLiteral(value)
        elif isinstance(value, bytearray):
            return BinaryLiteral(value)
        elif isinstance(value, Decimal):
            return DecimalLiteral(value)

    def to(self, type_var):
        raise NotImplementedError()

    def to_byte_buffer(self):
        raise NotImplementedError()


class BaseLiteral(Literal):
    def __init__(self, value, type_id):
        self.value = value
        self.byte_buffer = None
        self.type_id = type_id

    def to(self, type_var):
        raise NotImplementedError()

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, BaseLiteral):
            return False

        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "BaseLiteral(%s)" % str(self.value)

    def __str__(self):
        return str(self.value)

    def to_byte_buffer(self):
        if self.byte_buffer is None:
            self.byte_buffer = Conversions.to_byte_buffer(self.type_id, self.value)

        return self.byte_buffer


class ComparableLiteral(BaseLiteral):

    def __init__(self, value, type_id):
        super(ComparableLiteral, self).__init__(value, type_id)

    def to(self, type):
        raise NotImplementedError()

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if self.value is None:
            return True

        if other is None or other.value is None:
            return False

        return self.value < other.value

    def __gt__(self, other):
        if self.value is None:
            return False

        if other is None or other.value is None:
            return True

        return self.value > other.value

    def __le__(self, other):
        if self.value is None:
            return True

        if other is None or other.value is None:
            return False

        return self.value <= other.value

    def __ge__(self, other):
        if self.value is None:
            return False

        if other is None or other.value is None:
            return True

        return self.value >= other.value


class AboveMax(Literal):
    def __init__(self):
        super(AboveMax, self).__init__()

    def value(self):
        raise RuntimeError("AboveMax has no value")

    def to(self, type):
        raise RuntimeError("Cannot change the type of AboveMax")

    def __str__(self):
        return "aboveMax"


class BelowMin(Literal):
    def __init__(self):
        super(BelowMin, self).__init__()

    def value(self):
        raise RuntimeError("BelowMin has no value")

    def to(self, type):
        raise RuntimeError("Cannot change the type of BelowMin")

    def __str__(self):
        return "belowMin"


class BooleanLiteral(ComparableLiteral):

    def __init__(self, value):
        super(BooleanLiteral, self).__init__(value, TypeID.BOOLEAN)

    def to(self, type_var):
        if type_var.type_id == TypeID.BOOLEAN:
            return self


class IntegerLiteral(ComparableLiteral):

    def __init__(self, value):
        super(IntegerLiteral, self).__init__(value, TypeID.INTEGER)

    def to(self, type_var):
        if type_var.type_id == TypeID.INTEGER:
            return self
        elif type_var.type_id == TypeID.LONG:
            return LongLiteral(self.value)
        elif type_var.type_id == TypeID.FLOAT:
            return FloatLiteral(float(self.value))
        elif type_var.type_id == TypeID.DOUBLE:
            return DoubleLiteral(float(self.value))
        elif type_var.type_id == TypeID.DATE:
            return DateLiteral(self.value)
        elif type_var.type_id == TypeID.DECIMAL:
            if type_var.scale == 0:
                return DecimalLiteral(Decimal(self.value))
            else:
                return DecimalLiteral(Decimal(self.value)
                                      .quantize(Decimal("." + "".join(["0" for i in range(1, type_var.scale)]) + "1"),
                                                rounding=ROUND_HALF_UP))


class LongLiteral(ComparableLiteral):

    def __init__(self, value):
        super(LongLiteral, self).__init__(value, TypeID.LONG)

    def to(self, type_var):  # noqa: C901
        if type_var.type_id == TypeID.INTEGER:
            if Literal.JAVA_MAX_INT < self.value:
                return ABOVE_MAX
            elif Literal.JAVA_MIN_INT > self.value:
                return BELOW_MIN

            return IntegerLiteral(self.value)
        elif type_var.type_id == TypeID.LONG:
            return self
        elif type_var.type_id == TypeID.FLOAT:
            return FloatLiteral(float(self.value))
        elif type_var.type_id == TypeID.DOUBLE:
            return DoubleLiteral(float(self.value))
        elif type_var.type_id == TypeID.TIME:
            return TimeLiteral(self.value)
        elif type_var.type_id == TypeID.TIMESTAMP:
            return TimestampLiteral(self.value)
        elif type_var.type_id == TypeID.DECIMAL:
            if type_var.scale == 0:
                return DecimalLiteral(Decimal(self.value))
            else:
                return DecimalLiteral(Decimal(self.value)
                                      .quantize(Decimal("." + "".join(["0" for i in range(1, type_var.scale)]) + "1"),
                                                rounding=ROUND_HALF_UP))


class FloatLiteral(ComparableLiteral):

    def __init__(self, value):
        super(FloatLiteral, self).__init__(value, TypeID.FLOAT)

    def to(self, type_var):
        if type_var.type_id == TypeID.FLOAT:
            return self
        elif type_var.type_id == TypeID.DOUBLE:
            return DoubleLiteral(self.value)
        elif type_var.type_id == TypeID.DECIMAL:
            if type_var.scale == 0:
                return DecimalLiteral(Decimal(self.value)
                                      .quantize(Decimal('1.'),
                                                rounding=ROUND_HALF_UP))
            else:
                return DecimalLiteral(Decimal(self.value)
                                      .quantize(Decimal("." + "".join(["0" for i in range(1, type_var.scale)]) + "1"),
                                                rounding=ROUND_HALF_UP))


class DoubleLiteral(ComparableLiteral):

    def __init__(self, value):
        super(DoubleLiteral, self).__init__(value, TypeID.DOUBLE)

    def to(self, type_var):
        if type_var.type_id == TypeID.FLOAT:
            if JAVA_MAX_FLOAT < self.value:
                return ABOVE_MAX
            elif JAVA_MIN_FLOAT > self.value:
                return BELOW_MIN

            return FloatLiteral(self.value)
        elif type_var.type_id == TypeID.DOUBLE:
            return self
        elif type_var.type_id == TypeID.DECIMAL:
            if type_var.scale == 0:
                return DecimalLiteral(Decimal(self.value)
                                      .quantize(Decimal('1.'),
                                                rounding=ROUND_HALF_UP))
            else:
                return DecimalLiteral(Decimal(self.value)
                                      .quantize(Decimal("." + "".join(["0" for i in range(1, type_var.scale)]) + "1"),
                                                rounding=ROUND_HALF_UP))


class DateLiteral(ComparableLiteral):

    def __init__(self, value):
        super(DateLiteral, self).__init__(value, TypeID.DATE)

    def to(self, type_var):
        if type_var.type_id == TypeID.DATE:
            return self


class TimeLiteral(ComparableLiteral):

    def __init__(self, value):
        super(TimeLiteral, self).__init__(value, TypeID.TIME)

    def to(self, type_var):
        if type_var.type_id == TypeID.TIME:
            return self


class TimestampLiteral(ComparableLiteral):

    def __init__(self, value):
        super(TimestampLiteral, self).__init__(value, TypeID.TIMESTAMP)

    def to(self, type_var):
        if type_var.type_id == TypeID.TIMESTAMP:
            return self
        elif type_var.type_id == TypeID.DATE:
            return DateLiteral((datetime.datetime.fromtimestamp(self.value / 1000000) - Literals.EPOCH).days)


class DecimalLiteral(ComparableLiteral):

    def __init__(self, value):
        super(DecimalLiteral, self).__init__(value, TypeID.DECIMAL)

    def to(self, type_var):
        if type_var.type_id == TypeID.DECIMAL and type_var.scale == abs(self.value.as_tuple().exponent):
            return self


class StringLiteral(BaseLiteral):
    def __init__(self, value):
        super(StringLiteral, self).__init__(value, TypeID.STRING)

    def to(self, type_var):  # noqa: C901
        import dateutil.parser
        if type_var.type_id == TypeID.DATE:
            return DateLiteral((dateutil.parser.parse(self.value) - Literals.EPOCH).days)
        elif type_var.type_id == TypeID.TIME:
            return TimeLiteral(
                int((dateutil.parser.parse(Literals.EPOCH.strftime("%Y-%m-%d ") + self.value) - Literals.EPOCH)
                    .total_seconds() * 1000000))
        elif type_var.type_id == TypeID.TIMESTAMP:
            timestamp = dateutil.parser.parse(self.value)
            EPOCH = Literals.EPOCH
            if bool(timestamp.tzinfo) != bool(type_var.adjust_to_utc):
                raise RuntimeError("Cannot convert to %s when string is: %s" % (type_var, self.value))

            if timestamp.tzinfo is not None:
                EPOCH = EPOCH.replace(tzinfo=pytz.UTC)

            return TimestampLiteral(int((timestamp - EPOCH).total_seconds() * 1000000))
        elif type_var.type_id == TypeID.STRING:
            return self
        elif type_var.type_id == TypeID.UUID:
            return UUIDLiteral(uuid.UUID(self.value))
        elif type_var.type_id == TypeID.DECIMAL:
            dec_val = Decimal(str(self.value))
            if abs(dec_val.as_tuple().exponent) == type_var.scale:
                if type_var.scale == 0:
                    return DecimalLiteral(Decimal(str(self.value))
                                          .quantize(Decimal('1.'),
                                                    rounding=ROUND_HALF_UP))
                else:
                    return DecimalLiteral(Decimal(str(self.value))
                                          .quantize(Decimal("." + "".join(["0" for i in range(1, type_var.scale)]) + "1"),
                                                    rounding=ROUND_HALF_UP))

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, StringLiteral):
            return False

        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if other is None:
            return False

        return self.value < other.value

    def __gt__(self, other):
        if other is None:
            return True

        return self.value > other.value

    def __le__(self, other):
        if other is None:
            return False

        return self.value <= other.value

    def __ge__(self, other):
        if other is None:
            return True

        return self.value >= other.value

    def __str__(self):
        return '"' + self.value + '"'


class UUIDLiteral(ComparableLiteral):
    def __init__(self, value):
        super(UUIDLiteral, self).__init__(value, TypeID.UUID)

    def to(self, type_var):
        if type_var.type_id == TypeID.UUID:
            return self


class FixedLiteral(BaseLiteral):
    def __init__(self, value):
        super(FixedLiteral, self).__init__(value, TypeID.FIXED)

    def to(self, type_var):
        if type_var.type_id == TypeID.FIXED:
            if len(self.value) == type_var.length:
                return self
        elif type_var.type_id == TypeID.BINARY:
            return BinaryLiteral(self.value)

    def write_replace(self):
        return FixedLiteralProxy(self.value)

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if other is None:
            return False

        return self.value < other.value

    def __gt__(self, other):
        if other is None:
            return True

        return self.value > other.value

    def __le__(self, other):
        if other is None:
            return False

        return self.value <= other.value

    def __ge__(self, other):
        if other is None:
            return True

        return self.value >= other.value


class BinaryLiteral(BaseLiteral):
    def __init__(self, value):
        super(BinaryLiteral, self).__init__(value, TypeID.BINARY)

    def to(self, type_var):
        if type_var.type_id == TypeID.FIXED:
            if type_var.length == len(self.value):
                return FixedLiteral(self.value)
            return None
        elif type_var.type_id == TypeID.BINARY:
            return self

    def write_replace(self):
        return BinaryLiteralProxy(self.value)

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if other is None:
            return False

        return self.value < other.value

    def __gt__(self, other):
        if other is None:
            return True

        return self.value > other.value

    def __le__(self, other):
        if other is None:
            return False

        return self.value <= other.value

    def __ge__(self, other):
        if other is None:
            return True

        return self.value >= other.value


class FixedLiteralProxy(object):

    def __init__(self, buffer=None):
        if buffer is not None:
            self.bytes = list(buffer)

    def read_resolve(self):
        return FixedLiteral(self.bytes)


class ConstantExpressionProxy(object):

    def __init__(self, true_or_false=None):
        if true_or_false is not None:
            self.true_or_false = true_or_false

    def read_resolve(self):
        if self.true_or_false:
            return TRUE
        else:
            return FALSE


class BinaryLiteralProxy(FixedLiteralProxy):

    def __init__(self, buffer=None):
        super(BinaryLiteralProxy, self).__init__(buffer)

    def read_resolve(self):
        return BinaryLiteral(self.bytes)


ABOVE_MAX = AboveMax()
BELOW_MIN = BelowMin()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .transform import Transform


class VoidTransform(Transform):
    _INSTANCE = None

    @staticmethod
    def get():
        if VoidTransform._INSTANCE is None:
            VoidTransform._INSTANCE = VoidTransform()
        return VoidTransform._INSTANCE

    def __init__(self):
        pass

    def apply(self, value):
        return None

    def can_transform(self, type_var):
        return True

    def get_result_type(self, source_type):
        return source_type

    def project(self, name, predicate):
        return None

    def project_strict(self, name, predicate):
        return None

    def to_human_string(self, value):
        return "null"

    def __str__(self):
        return "void"
